# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Readers for ASCAT Level 1b and Level 2 data in EPS Native format.
"""

import os
import fnmatch
from gzip import GzipFile
from collections import OrderedDict, defaultdict
from tempfile import NamedTemporaryFile

import numpy as np
import xarray as xr
import lxml.etree as etree
from cadati.jd_date import jd2dt
from datetime import datetime
from datetime import timedelta

from ascat.utils import get_toi_subset, get_roi_subset
from ascat.utils import get_bit, set_bit
from ascat.utils import dtype_to_nan
from ascat.utils import mask_dtype_nans
from ascat.utils import int8_nan, uint8_nan
from ascat.utils import int16_nan, uint16_nan
from ascat.utils import int32_nan, uint32_nan
from ascat.utils import float32_nan
from ascat.read_native import AscatFile

short_cds_time = np.dtype([("day", ">u2"), ("time", ">u4")])
long_cds_time = np.dtype([("day", ">u2"), ("ms", ">u4"), ("mms", ">u2")])


long_nan = int32_nan
ulong_nan = uint32_nan
int_nan = int16_nan
uint_nan = uint16_nan

# 2000-01-01 00:00:00
julian_epoch = 2451544.5


class AscatL1bEpsSzfFile(AscatFile):
    """
    Class reading ASCAT Level 1b file in EPS Native format.
    """

    def _read(self, filename, toi=None, roi=None, generic=True, to_xarray=False,
             ignore_noise_ool=False):
        """
        Read one ASCAT Level 1b EPS Szf file.

        Parameters
        ----------
        toi : tuple of datetime, optional
            Filter data for given time of interest (default: None).
            e.g. (datetime(2020, 1, 1, 12), datetime(2020, 1, 2))
        roi : tuple of 4 float, optional
            Filter data for region of interest (default: None).
            e.g. latmin, lonmin, latmax, lonmax
        generic : boolean, optional
            Convert original data field names to generic field names
            (default: True).
        to_xarray : boolean, optional
            Convert data to xarray.Dataset otherwise numpy.ndarray will be
            returned (default: False).
        ignore_noise_ool : bool, optional
            Ignore noise out of limit flag (default: False).

        Returns
        -------
        data : xarray.Dataset or numpy.ndarray
            ASCAT data.
        metadata : dict
            Metadata.

        Notes
        -----
        TODO Decide whether to do subsetting here (per file) or later
        (after merging). At the moment the possibility is here but it is not used
        by super().read()
        """
        data, metadata = read_eps_l1b(
            filename,
            generic,
            to_xarray,
            full=False,
            unsafe=True,
            scale_mdr=False,
            ignore_noise_ool=ignore_noise_ool)

        if toi:
            data = get_toi_subset(data, toi)

        if roi:
            data = get_roi_subset(data, roi)

        return data, metadata

    def _merge(self, data):
        """
        Merge data.

        Parameters
        ----------
        data : list
            List of array.

        Returns
        -------
        data : numpy.ndarray
            Data.
        """
        metadata = {}

        left_beams = ["lf-vv", "lm-vv", "la-vv"]
        right_beams = ["rf-vv", "rm-vv", "ra-vv"]
        all_beams = left_beams + right_beams

        if isinstance(data[0], tuple):
            data, metadata = zip(*data)

        merged_data = defaultdict(list)
        for beam in all_beams:
            for d in data:
                merged_data[beam].append(d.pop(beam))
            if isinstance(merged_data[beam][0], xr.Dataset):
                merged_data[beam] = xr.concat(merged_data[beam],
                                              dim="obs",
                                              combine_attrs="drop_conflicts")
            else:
                merged_data[beam] = np.hstack(merged_data[beam])

        merged_data = (merged_data, metadata)

        return merged_data

class AscatL1bEpsFile(AscatFile):
    """
    ASCAT Level 1b EPS Native reader class.
    """

    def _read(self, filename, generic=False, to_xarray=False, **kwargs):
        """
        Read one ASCAT Level 1b EPS file.

        Parameters
        ----------
        generic : boolean, optional
            Convert original data field names to generic field names
            (default: False).
        to_xarray : boolean, optional
            Convert data to xarray.Dataset otherwise numpy.ndarray will be
            returned (default: False).

        Returns
        -------
        data : xarray.Dataset or numpy.ndarray
            ASCAT data.
        metadata : dict
            Metadata.
        """
        return read_eps_l1b(filename, generic, to_xarray, return_ptype=True, **kwargs)

    def _merge(self, data):
        """
        Merge data.

        Parameters
        ----------
        data : list
            List of array.

        Returns
        -------
        data : xarray.Dataset or numpy.ndarray
            Data.
        """
        ptype = data[0][1]["product_type"]
        metadata = {}

        left_beams = ["lf-vv", "lm-vv", "la-vv"]
        right_beams = ["rf-vv", "rm-vv", "ra-vv"]
        all_beams = left_beams + right_beams

        if isinstance(data[0], tuple):
            data, metadata = zip(*data)
            if ptype == "szf":
                merged_data = defaultdict(list)
                for beam in all_beams:
                    for d in data:
                        merged_data[beam].append(d.pop(beam))
                    if isinstance(merged_data[beam][0], xr.Dataset):
                        merged_data[beam] = xr.concat(merged_data[beam],
                                                      dim="obs",
                                                      combine_attrs="drop_conflicts")
                    else:
                        merged_data[beam] = np.hstack(merged_data[beam])
            else:
                if isinstance(merged_data[beam][0], xr.Dataset):
                    merged_data = xr.concat(data, dim="obs", combine_attrs="drop_conflicts")
                else:
                    merged_data = np.hstack(data)

        # if ptype == "szf":
        #     if isinstance(data[0], tuple):
        #         data, metadata = zip(*data)
        #     merged_data = defaultdict(list)
        #     for beam in all_beams:
        #         for d in data:
        #             merged_data[beam].append(d.pop(beam))
        #         merged_data[beam] = np.hstack(merged_data[beam])
        # else:
        #     if isinstance(data[0], tuple):
        #         data, metadata = zip(*data)
        #     merged_data = np.hstack(data)

        merged_data = (merged_data, metadata)

        return merged_data

class AscatL1bEpsFileGeneric(AscatL1bEpsFile):
    """
    The same as AscatL1bEpsFile but with generic=True by default.
    """
    def _read(self, filename, generic=True, to_xarray=False, **kwargs):
        return super()._read(filename, generic=generic, to_xarray=to_xarray, **kwargs)


class AscatL2EpsFile(AscatFile):
    """
    ASCAT Level 2 EPS Native reader class.
    """

    def _read(self, filename, generic=False, to_xarray=False, **kwargs):
        """
        Read one ASCAT Level 2 EPS file.

        Returns
        -------
        generic : boolean, optional
            Convert original data field names to generic field names
            (default: False).
        to_xarray : boolean, optional
            Convert data to xarray.Dataset otherwise numpy.ndarray will be
            returned (default: False).

        Returns
        -------
        data : xarray.Dataset or numpy.ndarray
            ASCAT data.
        metadata : dict
            Metadata.
        """
        return read_eps_l2(filename, generic=generic, to_xarray=to_xarray, **kwargs)

    def _merge(self, data):
        """
        Merge data.

        Parameters
        ----------
        data : list
            List of array.

        Returns
        -------
        data : numpy.ndarray
            Data.
        """
        if isinstance(data[0], tuple):
            data, metadata = zip(*data)
            if isinstance(data[0], xr.Dataset):
                data = xr.concat(data, dim="obs", combine_attrs="drop_conflicts")
            else:
                data = np.hstack(data)
            data = (data, metadata)
        else:
            data = np.hstack(data)

        return data

class AscatL2EpsFileGeneric(AscatL2EpsFile):
    """
    The same as AscatL1bEpsFile but with generic=True by default.
    """
    def _read(self, filename, generic=True, to_xarray=False, **kwargs):
        return super()._read(filename, generic=generic, to_xarray=to_xarray, **kwargs)

class EPSProduct:
    """
    Class for reading EPS products.
    """

    def __init__(self, filename):
        """
        Initialize EPSProduct.

        Parameters
        ----------
        filename : str
            EPS Native Filename.
        """
        self.filename = filename
        self.fid = None
        self.mphr = None
        self.sphr = None
        self.aux = defaultdict(list)
        self.mdr = None
        self.scaled_mdr = None
        self.xml_file = None
        self.xml_doc = None
        self.mdr_template = None
        self.scaled_template = None
        self.sfactor = None

        self.grh_dtype = np.dtype([("record_class", "u1"),
                                   ("instrument_group", "u1"),
                                   ("record_subclass", "u1"),
                                   ("record_subclass_version", "u1"),
                                   ("record_size", ">u4"),
                                   ("record_start_time", short_cds_time),
                                   ("record_stop_time", short_cds_time)])

        self.ipr_dtype = np.dtype([("grh", self.grh_dtype),
                                   ("target_record_class", "u1"),
                                   ("target_instrument_group", "u1"),
                                   ("target_record_subclass", "u1"),
                                   ("target_record_offset", ">u4")])

        self.pointer_dtype = np.dtype([("grh", self.grh_dtype),
                                       ("aux_data_pointer", "u1", 100)])

        self.filesize = os.path.getsize(self.filename)

    def read_mphr(self):
        """
        Read only Main Product Header Record (MPHR).
        """
        with open(self.filename, "rb") as fid:
            grh = np.fromfile(fid, dtype=self.grh_dtype, count=1)[0]
            if grh["record_class"] == 1:
                mphr = fid.read(grh["record_size"] - grh.itemsize)
                mphr = OrderedDict(
                    item.replace(" ", "").split("=")
                    for item in mphr.decode("utf-8").split("\n")[:-1])

        return mphr

    def read(self, full=True, unsafe=False, scale_mdr=True):
        """
        Read EPS file.

        Parameters
        ----------
        full : bool, optional
            Read full file content (True) or just Main Product Header
            Record (MPHR) and Main Data Record (MDR) (False). Default: True
        unsafe : bool, optional
            If True it is (unsafely) assumed that MDR are continuously
            stacked until the end of file. Makes reading a lot faster.
            Default: False
        scale_mdr : bool, optional
            Compute scaled MDR (True) or not (False). Default: True

        Returns
        -------
        mphr : dict self.sphr, self.aux, self.mdr, scaled_mdr
            Main Product Header Record (MPHR).
        sphr : dict
            Secondary Product Header Product (SPHR).
        aux : dict
            Auxiliary Header Products.
        mdr : numpy.ndarray
            Main Data Record (MDR)
        scaled_mdr : numpy.ndarray
            Scaled Main Data Record (MPHR) or None if not computed.
        """
        self.fid = open(self.filename, "rb")

        abs_pos = 0
        grh = None
        prev_grh = None
        record_count = 0

        start_dt = datetime.now()

        while True:
            # read generic record header of data block
            grh = np.fromfile(self.fid, dtype=self.grh_dtype, count=1)[0]

            if grh["record_class"] == 8 and unsafe:
                if np.mod((self.filesize - abs_pos),
                          self.mdr_template.itemsize) != 0:
                    # Unsafe reading fails, switching to safe reading
                    unsafe = False
                else:
                    num_mdr = (self.filesize -
                               abs_pos) // self.mdr_template.itemsize
                    self.fid.seek(abs_pos)
                    self.read_record_class(grh, num_mdr)
                    break

            if prev_grh is None:
                prev_grh = grh

            if ((prev_grh["record_class"] != grh["record_class"]) or
                (prev_grh["record_subclass"] != grh["record_subclass"])):

                # compute record start position of previous record
                start_pos = (abs_pos - prev_grh["record_size"] * record_count)
                self.fid.seek(start_pos)

                if full or (prev_grh["record_class"] == 8 or
                            prev_grh["record_class"] == 1):
                    # read previous record, because new one is coming
                    self.read_record_class(prev_grh, record_count)

                # reset record class count
                record_count = 1
            else:
                # same record class as before, increase count
                record_count += 1

            abs_pos += grh["record_size"]

            # position after record
            self.fid.seek(abs_pos)

            # store grh
            prev_grh = grh

            # end of file?
            if abs_pos == self.filesize:

                # compute record start position of previous record class
                start_pos = (abs_pos - prev_grh["record_size"] * record_count)
                self.fid.seek(start_pos)

                # read final record class(es)
                self.read_record_class(prev_grh, record_count)

                break

            if (datetime.now() - start_dt) > timedelta(minutes=3):
                print("Timeout reading EPS file")
                self.mdr = None
                break

        self.fid.close()

        if scale_mdr:
            self.scaled_mdr = self._scaling(self.mdr, self.scaled_template,
                                            self.mdr_sfactor)

        return self.mphr, self.sphr, self.aux, self.mdr, self.scaled_mdr

    def read_record_class(self, grh, record_count):
        """
        Read record class.

        Parameters
        ----------
        grh : numpy.ndarray
            Generic record header.
        record_count : int
            Number of records.
        """
        # mphr (Main Product Header Reader)
        if grh["record_class"] == 1:
            self.fid.seek(grh.itemsize, 1)
            self._read_mphr(grh)

            # find the xml file corresponding to the format version
            # and load template
            self.xml_file = self._get_eps_xml()
            self.xml_doc = etree.parse(self.xml_file)
            self.mdr_template, self.scaled_template, self.mdr_sfactor = \
                self._read_xml_mdr()

        # sphr (Secondary Product Header Record)
        elif grh["record_class"] == 2:
            self.fid.seek(grh.itemsize, 1)
            self._read_sphr(grh)

        # ipr (Internal Pointer Record)
        elif grh["record_class"] == 3:
            data = np.fromfile(
                self.fid, dtype=self.ipr_dtype, count=record_count)
            self.aux["ipr"].append(data)

        # geadr (Global External Auxiliary Data Record)
        elif grh["record_class"] == 4:
            data = self._read_pointer(record_count)
            self.aux["geadr"].append(data)

        # veadr (Variable External Auxiliary Data Record)
        elif grh["record_class"] == 6:
            data = self._read_pointer(record_count)
            self.aux["veadr"].append(data)

        # viadr (Variable Internal Auxiliary Data Record)
        elif grh["record_class"] == 7:
            template, scaled_template, sfactor = self._read_xml_viadr(
                grh["record_subclass"])
            viadr_element = np.fromfile(
                self.fid, dtype=template, count=record_count)

            viadr_element_sc = self._scaling(viadr_element, scaled_template,
                                             sfactor)

            # store viadr_grid separately
            if grh["record_subclass"] == 8:
                self.aux["viadr_grid"].append(viadr_element)
                self.aux["viadr_grid_scaled"].append(viadr_element_sc)
            else:
                self.aux["viadr"].append(viadr_element)
                self.aux["viadr_scaled"].append(viadr_element_sc)

        # mdr (Measurement Data Record)
        elif grh["record_class"] == 8:
            if grh["instrument_group"] == 13:
                self.dummy_mdr = np.fromfile(
                    self.fid, dtype=self.mdr_template, count=record_count)
            else:
                self.mdr = np.fromfile(
                    self.fid, dtype=self.mdr_template, count=record_count)
                self.mdr_counter = record_count
        else:
            raise RuntimeError("Record class not found.")

    def _scaling(self, unscaled_mdr, scaled_template, sfactor):
        """
        Scale the MDR.

        Parameters
        ----------
        unscaled_mdr : numpy.ndarray
            Raw MDR.
        scaled_template : numpy.dtype
            Scaled MDR template.
        sfactor : dict
            Scale factors.

        Returns
        -------
        scaled_mdr : numpy.ndarray
            Scaled MDR.
        """
        scaled_mdr = np.empty(unscaled_mdr.shape, dtype=scaled_template)

        for key, value in sfactor.items():
            if value != 1:
                scaled_mdr[key] = unscaled_mdr[key] * 1. / value
            else:
                scaled_mdr[key] = unscaled_mdr[key]

        return scaled_mdr

    def _read_mphr(self, grh):
        """
        Read Main Product Header (MPHR).
        """
        mphr = self.fid.read(grh["record_size"] - grh.itemsize)
        self.mphr = OrderedDict(
            item.replace(" ", "").split("=")
            for item in mphr.decode("utf-8").split("\n")[:-1])

    def _read_sphr(self, grh):
        """
        Read Special Product Header (SPHR).
        """
        sphr = self.fid.read(grh["record_size"] - grh.itemsize)
        self.sphr = OrderedDict(
            item.replace(" ", "").split("=")
            for item in sphr.decode("utf-8").split("\n")[:-1])

    def _read_pointer(self, count=1):
        """
        Read pointer record.
        """
        record = np.fromfile(self.fid, dtype=self.pointer_dtype, count=count)

        return record

    def _get_eps_xml(self):
        """
        Find the corresponding eps xml file.
        """
        format_path = os.path.join(os.path.dirname(__file__), "formats")

        # loop through files where filename starts with "eps_ascat".
        for filename in fnmatch.filter(os.listdir(format_path), "eps_ascat*"):
            doc = etree.parse(os.path.join(format_path, filename))
            file_extension = doc.xpath("//file-extensions")[0].getchildren()[0]

            format_version = doc.xpath("//format-version")
            for elem in format_version:
                major = elem.getchildren()[0]
                minor = elem.getchildren()[1]

                # return the xml file matching the metadata of the datafile.
                if major.text == self.mphr["FORMAT_MAJOR_VERSION"] and \
                        minor.text == self.mphr["FORMAT_MINOR_VERSION"] and \
                        self.mphr[
                            "PROCESSING_LEVEL"] in file_extension.text and \
                        self.mphr["PRODUCT_TYPE"] in file_extension.text:
                    return os.path.join(format_path, filename)

    def _read_xml_viadr(self, subclassid):
        """
        Read xml record of viadr class.
        """
        elements = self.xml_doc.xpath("//viadr")
        data = OrderedDict()
        length = []

        # find the element with the correct subclass
        for elem in elements:
            item_dict = dict(elem.items())
            subclass = int(item_dict["subclass"])
            if subclass == subclassid:
                break

        for child in elem.getchildren():

            if child.tag == "delimiter":
                continue

            child_items = dict(child.items())
            name = child_items.pop("name")

            # check if the item is of type longtime
            longtime_flag = ("type" in child_items and
                             "longtime" in child_items["type"])

            # append the length if it isn"t the special case of type longtime
            try:
                var_len = child_items.pop("length")
                if not longtime_flag:
                    length.append(np.int64(var_len))
            except KeyError:
                pass

            data[name] = child_items

            if child.tag == "array":
                for arr in child.iterdescendants():
                    arr_items = dict(arr.items())
                    if arr.tag == "field":
                        data[name].update(arr_items)
                    else:
                        try:
                            var_len = arr_items.pop("length")
                            length.append(np.int64(var_len))
                        except KeyError:
                            pass

            if length:
                data[name].update({"length": length})
            else:
                data[name].update({"length": 1})

            length = []

        conv = {
            "longtime": long_cds_time,
            "time": short_cds_time,
            "boolean": "u1",
            "integer1": "i1",
            "uinteger1": "u1",
            "integer": ">i4",
            "uinteger": ">u4",
            "integer2": ">i2",
            "uinteger2": ">u2",
            "integer4": ">i4",
            "uinteger4": ">u4",
            "integer8": ">i8",
            "enumerated": "u1",
            "string": "str",
            "bitfield": "u1"
        }

        scaling_factor = {}
        scaled_dtype = []
        dtype = []

        for key, value in data.items():

            if "scaling-factor" in value:
                sf_dtype = np.float32
                sf_split = value["scaling-factor"].split("^")
                scaling_factor[key] = np.int64(sf_split[0])**np.int64(
                    sf_split[1])
            else:
                sf_dtype = conv[value["type"]]
                scaling_factor[key] = 1

            length = value["length"]

            if length == 1:
                scaled_dtype.append((key, sf_dtype))
                dtype.append((key, conv[value["type"]]))
            else:
                scaled_dtype.append((key, sf_dtype, length))
                dtype.append((key, conv[value["type"]], length))

        return np.dtype(dtype), np.dtype(scaled_dtype), scaling_factor

    def _read_xml_mdr(self):
        """
        Read xml record of mdr class.
        """
        elements = self.xml_doc.xpath("//mdr")
        data = OrderedDict()
        length = []
        elem = elements[0]

        for child in elem.getchildren():

            if child.tag == "delimiter":
                continue

            child_items = dict(child.items())
            name = child_items.pop("name")

            # check if the item is of type bitfield
            bitfield_flag = ("type" in child_items and
                             ("bitfield" in child_items["type"] or
                              "time" in child_items["type"]))

            # append the length if it isn"t the special case of type
            # bitfield or time
            try:
                var_len = child_items.pop("length")
                if not bitfield_flag:
                    length.append(np.int64(var_len))
            except KeyError:
                pass

            data[name] = child_items

            if child.tag == "array":
                for arr in child.iterdescendants():
                    arr_items = dict(arr.items())

                    # check if the type is bitfield
                    bitfield_flag = ("type" in arr_items and
                                     "bitfield" in arr_items["type"])

                    if bitfield_flag:
                        data[name].update(arr_items)
                        break
                    else:
                        if arr.tag == "field":
                            data[name].update(arr_items)
                        else:
                            try:
                                var_len = arr_items.pop("length")
                                length.append(np.int64(var_len))
                            except KeyError:
                                pass

            if length:
                data[name].update({"length": length})
            else:
                data[name].update({"length": 1})

            length = []

        conv = {
            "longtime": long_cds_time,
            "time": short_cds_time,
            "boolean": "u1",
            "integer1": "i1",
            "uinteger1": "u1",
            "integer": ">i4",
            "uinteger": ">u4",
            "integer2": ">i2",
            "uinteger2": ">u2",
            "integer4": ">i4",
            "uinteger4": ">u4",
            "integer8": ">i8",
            "enumerated": "u1",
            "string": "str",
            "bitfield": "u1"
        }

        scaling_factor = {}
        scaled_dtype = []
        dtype = [("grh", self.grh_dtype)]

        for key, value in data.items():

            if "scaling-factor" in value:
                sf_dtype = np.float32
                sf_split = value["scaling-factor"].split("^")
                scaling_factor[key] = np.int64(sf_split[0])**np.int64(
                    sf_split[1])
            else:
                sf_dtype = conv[value["type"]]
                scaling_factor[key] = 1

            length = value["length"]

            if length == 1:
                scaled_dtype.append((key, sf_dtype))
                dtype.append((key, conv[value["type"]]))
            else:
                scaled_dtype.append((key, sf_dtype, length))
                dtype.append((key, conv[value["type"]], length))

        return np.dtype(dtype), np.dtype(scaled_dtype), scaling_factor


def conv_epsl1bszf_generic(data, metadata, gen_fields_lut, skip_fields):
    """
    Rename and convert data types of dataset.

    Parameters
    ----------
    data : dict of numpy.ndarray
        Original dataset.
    metadata : dict
        Metadata.

    Returns
    -------
    data : dict of numpy.ndarray
        Converted dataset.
    """
    for var_name in skip_fields:
        data.pop(var_name, None)

    for var_name, (new_name, new_dtype, valid_range,
                   nan_val) in gen_fields_lut.items():
        if new_dtype is None:
            data[new_name] = np.ma.array(data.pop(var_name))
            data[new_name].mask = ((data[new_name] < valid_range[0]) |
                                   (data[new_name] > valid_range[1]))
            data[new_name].set_fill_value(nan_val)
        else:
            invalid = data[var_name] == dtype_to_nan[np.dtype(data[var_name].dtype)]
            data[new_name] = np.ma.array(data.pop(var_name).astype(new_dtype))
            data[new_name].mask = ((data[new_name] < valid_range[0]) |
                                   (data[new_name] > valid_range[1]) |
                                   invalid)
            data[new_name].set_fill_value(nan_val)

    return data


def conv_epsl1bszx_generic(data, metadata):
    """
    Rename and convert data types of dataset.

    Parameters
    ----------
    data : dict of numpy.ndarray
        Original dataset.
    metadata : dict
        Metadata.

    Returns
    -------
    data : dict of numpy.ndarray
        Converted dataset.
    """
    # template - "old_var_name": ("new_name", new dtype )
    gen_fields_lut = {
        "inc_angle_trip": ("inc", np.float32, uint_nan),
        "azi_angle_trip": ("azi", np.float32, int_nan),
        "sigma0_trip": ("sig", np.float32, long_nan),
        "kp": ("kp", np.float32, uint_nan),
        "f_kp": ("kp_quality", np.uint8, None), # "f_kp": ("kp_quality", np.int8, uint8_nan),
        "f_usable": ("f_usable", np.int8, uint8_nan),
        "swath_indicator": ("swath_indicator", np.int8, uint8_nan),
    }

    skip_fields = ["flagfield_rf1", "f_f", "f_v", "f_oa", "f_sa", "f_tel"]

    for var_name in skip_fields:
        if var_name in data:
            data.pop(var_name)

    for var_name, (new_name, new_dtype, nan_val) in gen_fields_lut.items():
        invalid = data[var_name] == nan_val
        data[new_name] = data.pop(var_name).astype(new_dtype)
        if nan_val is not None:
            new_nan_val = dtype_to_nan[np.dtype(new_dtype)]
            data[new_name][invalid] = new_nan_val

    data["sat_id"] = np.repeat(metadata["sat_id"], data["time"].size)

    return data


def conv_epsl2szx_generic(data, metadata):
    """
    Rename and convert data types of dataset.

    Parameters
    ----------
    data : dict of numpy.ndarray
        Original dataset.
    metadata : dict
        Metadata.

    Returns
    -------
    data : dict of numpy.ndarray
        Converted dataset.
    """
    gen_fields_lut = {
        "inc_angle_trip": ("inc", np.float32, uint_nan),
        "azi_angle_trip": ("azi", np.float32, int_nan),
        "sigma0_trip": ("sig", np.float32, long_nan),
        "soil_moisture": ("sm", np.float32, uint_nan),
        "soil_moisture_error": ("sm_noise", np.float32, uint_nan),
        "mean_surf_soil_moisture": ("sm_mean", np.float32, uint_nan),
        "soil_moisture_sensetivity": ("sm_sens", np.float32, ulong_nan),
        "sigma40": ("sig40", np.float32, long_nan),
        "sigma40_error": ("sig40_noise", np.float32, long_nan),
        "slope40": ("slope40", np.float32, long_nan),
        "slope40_error": ("slope40_noise", np.float32, long_nan),
        "dry_backscatter": ("dry_sig40", np.float32, long_nan),
        "wet_backscatter": ("wet_sig40", np.float32, long_nan),
        "as_des_pass": ("as_des_pass", np.uint8, None),
        "aggregated_quality_flag": ("agg_flag", np.uint8, None),
        "processing_flags": ("proc_flag", np.uint8, None),
        "correction_flags": ("corr_flag", np.uint8, None),
        "snow_cover_probability": ("snow_prob", np.uint8, None),
        "frozen_soil_probability": ("frozen_prob", np.uint8, None),
        "innudation_or_wetland": ("wetland", np.uint8, None),
        "topographical_complexity": ("topo", np.uint8, None),
        "kp": ("kp", np.float32, uint_nan),
        "swath_indicator": ("swath_indicator", np.int8, uint8_nan)
    }

    skip_fields = ["flagfield_rf1", "f_f", "f_v", "f_oa", "f_sa", "f_tel"]

    for var_name in skip_fields:
        if var_name in data:
            data.pop(var_name)

    for var_name, (new_name, new_dtype, nan_val) in gen_fields_lut.items():
        invalid = data[var_name] == nan_val
        data[new_name] = data.pop(var_name).astype(new_dtype)
        if nan_val is not None:
            new_nan_val = dtype_to_nan[np.dtype(new_dtype)]
            data[new_name][invalid] = new_nan_val

    data["sat_id"] = np.repeat(metadata["sat_id"], data["time"].size)

    return data


def read_eps_l1b(filename,
                 generic=False,
                 to_xarray=False,
                 full=True,
                 unsafe=False,
                 scale_mdr=True,
                 ignore_noise_ool=False,
                 return_ptype=False):
    """
    Level 1b reader and data preparation.

    Parameters
    ----------
    filename : str
        ASCAT Level 1b file name in EPS Native format.
    generic : bool, optional
        "True" reading and converting into generic format or
        "False" reading original field names (default: False).
    to_xarray : bool, optional
        "True" return data as xarray.Dataset
        "False" return data as numpy.ndarray (default: False).
    full : bool, optional
        Read full file content (True) or just Main Product Header
        Record (MPHR) and Main Data Record (MDR) (False). Default: True
    unsafe : bool, optional
        If True it is (unsafely) assumed that MDR are continuously
        stacked until the end of file. Makes reading a lot faster.
        Default: False
    scale_mdr : bool, optional
        Compute scaled MDR (True) or not (False). Default: True
    ignore_noise_ool : bool, optional
        Ignore noise out of limit flag (default: False).

    Returns
    -------
    ds : xarray.Dataset, dict of xarray.Dataset
        ASCAT Level 1b data.
    """
    eps_file = read_eps(
        filename, full=full, unsafe=unsafe, scale_mdr=scale_mdr)

    ptype = eps_file.mphr["PRODUCT_TYPE"]
    fmv = int(eps_file.mphr["FORMAT_MAJOR_VERSION"])

    if ptype == "SZF":

        if fmv == 12:
            data, metadata = read_szf_fmv_12(eps_file, ignore_noise_ool)

            skip_fields = [
                "utc_localisation-days", "utc_localisation-milliseconds",
                "degraded_inst_mdr", "degraded_proc_mdr", "flagfield_rf1",
                "flagfield_rf2", "flagfield_pl", "flagfield_gen1",
                "flagfield_gen2"
            ]

            gen_fields_lut = {
                "inc_angle_full": ("inc", np.float32, (0, 90), float32_nan),
                "azi_angle_full": ("azi", np.float32, (0, 360), float32_nan),
                "sigma0_full": ("sig", np.float32, (-50, 50), float32_nan),
                "sat_track_azi":
                    ("sat_track_azi", np.float32, (0, 360), float32_nan),
                "beam_number": ("beam_number", np.int8, (1, 6), int8_nan),
                "swath_indicator":
                    ("swath_indicator", np.int8, (0, 1), int8_nan),
                "land_frac": ("land_frac", np.float32, (0, 1), float32_nan),
                "f_usable": ("f_usable", np.int8, (0, 2), int8_nan),
                "as_des_pass": ("as_des_pass", np.uint8, (0, 1), uint8_nan),
                "time": ("time", None, (np.datetime64("1900-01-01"),
                                        np.datetime64("2100-01-01")), 0),
                "lon": ("lon", np.float32, (-180, 180), float32_nan),
                "lat": ("lat", np.float32, (-90, 90), float32_nan),
                "flagfield":
                    ("flagfield", None, (0, uint32_nan - 1), uint32_nan),
            }

        elif fmv == 13:
            data, metadata = read_szf_fmv_13(eps_file, ignore_noise_ool)

            skip_fields = [
                "utc_localisation-days",
                "utc_localisation-milliseconds",
                "degraded_inst_mdr",
                "degraded_proc_mdr",
            ]

            gen_fields_lut = {
                "inc_angle_full": ("inc", np.float32, (0, 90), float32_nan),
                "azi_angle_full": ("azi", np.float32, (0, 360), float32_nan),
                "sigma0_full": ("sig", np.float32, (-50, 50), float32_nan),
                "sat_track_azi":
                    ("sat_track_azi", np.float32, (0, 360), float32_nan),
                "beam_number": ("beam_number", np.int8, (1, 6), int8_nan),
                "swath_indicator":
                    ("swath_indicator", np.int8, (0, 1), int8_nan),
                # "land_frac": ("land_frac", np.float32, (0, 1), float32_nan),
                "f_usable": ("f_usable", np.int8, (0, 2), int8_nan),
                "as_des_pass": ("as_des_pass", np.uint8, (0, 1), uint8_nan),
                "time": ("time", None, (np.datetime64("1900-01-01"),
                                        np.datetime64("2100-01-01")), 0),
                "lon": ("lon", np.float32, (-180, 180), float32_nan),
                "lat": ("lat", np.float32, (-90, 90), float32_nan),
                "flagfield":
                    ("flagfield", None, (0, uint32_nan - 1), uint32_nan),
            }
        else:
            raise RuntimeError("L1b SZF format version not supported.")

        rename_coords = {"longitude_full": "lon", "latitude_full": "lat"}

        for k, v in rename_coords.items():
            data[v] = data.pop(k)

        if generic:
            data = conv_epsl1bszf_generic(data, metadata, gen_fields_lut,
                                          skip_fields)

        # 1 Left Fore Antenna, 2 Left Mid Antenna 3 Left Aft Antenna
        # 4 Right Fore Antenna, 5 Right Mid Antenna, 6 Right Aft Antenna
        left_beams = ["lf-vv", "lm-vv", "la-vv"]
        right_beams = ["rf-vv", "rm-vv", "ra-vv"]
        all_beams = left_beams + right_beams

        ds = OrderedDict()

        for i, beam in enumerate(all_beams):

            subset = data["beam_number"] == i + 1

            # convert spacecraft_id to internal sat_id
            sat_id = np.array([4, 3, 5])
            metadata["sat_id"] = sat_id[metadata["spacecraft_id"] - 1]

            # convert dict to xarray.Dataset or numpy.ndarray
            if to_xarray:
                sub_data = {}
                for var_name in data.keys():

                    if var_name == "beam_number" and generic:
                        continue

                    if len(data[var_name].shape) == 1:
                        dim = ["obs"]
                    elif len(data[var_name].shape) == 2:
                        dim = ["obs", "echo"]
                    if var_name == "time":
                        data[var_name] = data[var_name].astype("datetime64[ns]")

                    sub_data[var_name] = (dim, data[var_name][subset])

                coords = {}
                coords_fields = ["lon", "lat", "time"]

                for cf in coords_fields:
                    coords[cf] = sub_data.pop(cf)

                ds[beam] = xr.Dataset(sub_data, coords=coords, attrs=metadata)
                if generic:
                    data = mask_dtype_nans(data)
            else:
                # collect dtype info
                dtype = []
                fill_values = {}

                for var_name in data.keys():

                    if var_name == "beam_number" and generic:
                        continue

                    if len(data[var_name][subset].shape) == 1:
                        dtype.append(
                            (var_name, data[var_name][subset].dtype.str))
                    elif len(data[var_name][subset].shape) > 1:
                        dtype.append(
                            (var_name, data[var_name][subset].dtype.str,
                             data[var_name][subset].shape[1:]))

                    fill_values[var_name] = data[var_name].fill_value

                ds[beam] = np.ma.empty(
                    data["time"][subset].size, dtype=np.dtype(dtype))

                for var_name, v in data.items():
                    if var_name == "beam_number" and generic:
                        continue
                    ds[beam][var_name] = v[subset]
                    ds[beam][var_name].set_fill_value(fill_values[var_name])

    elif ptype in ["SZR", "SZO"]:

        if fmv == 11:
            data, metadata = read_szx_fmv_11(eps_file)
        elif fmv == 12:
            data, metadata = read_szx_fmv_12(eps_file)
        elif fmv == 13:
            data, metadata = read_szx_fmv_13(eps_file)
        else:
            raise RuntimeError("SZR/SZO format version not supported.")

        data["time"] = jd2dt(data.pop("jd"))

        rename_coords = {"longitude": "lon", "latitude": "lat"}

        for k, v in rename_coords.items():
            data[v] = data.pop(k)

        # convert spacecraft_id to internal sat_id
        sat_id = np.array([4, 3, 5])
        metadata["sat_id"] = sat_id[metadata["spacecraft_id"] - 1]

        # add/rename/remove fields according to generic format
        if generic:
            data = conv_epsl1bszx_generic(data, metadata)

        # convert dict to xarray.Dataset or numpy.ndarray
        if to_xarray:
            for k in data.keys():
                if len(data[k].shape) == 1:
                    dim = ["obs"]
                elif len(data[k].shape) == 2:
                    dim = ["obs", "beam"]
                if k == "time":
                    data[k] = data[k].astype("datetime64[ns]")

                data[k] = (dim, data[k])

            coords = {}
            coords_fields = ["lon", "lat", "time"]
            for cf in coords_fields:
                coords[cf] = data.pop(cf)

            ds = xr.Dataset(data, coords=coords, attrs=metadata)
            if generic:
                data = mask_dtype_nans(data)
        else:
            # collect dtype info
            dtype = []
            for var_name in data.keys():
                if len(data[var_name].shape) == 1:
                    dtype.append((var_name, data[var_name].dtype.str))
                elif len(data[var_name].shape) > 1:
                    dtype.append((var_name, data[var_name].dtype.str,
                                  data[var_name].shape[1:]))

            ds = np.empty(data["time"].size, dtype=np.dtype(dtype))
            for k, v in data.items():
                ds[k] = v
    else:
        raise RuntimeError("Format not supported. Product type {:1}"
                           " Format major version: {:2}".format(ptype, fmv))

    metadata["filename"] = os.path.basename(filename)
    if return_ptype:
        metadata["product_type"] = ptype

    return ds, metadata


def read_eps_l2(filename, generic=False, to_xarray=False, return_ptype=False):
    """
    Level 2 reader and data preparation.

    Parameters
    ----------
    filename : str
        ASCAT Level 1b file name in EPS Native format.
    generic : bool, optional
        "True" reading and converting into generic format or
        "False" reading original field names (default: False).
    to_xarray : bool, optional
        "True" return data as xarray.Dataset
        "False" return data as numpy.ndarray (default: False).

    Returns
    -------
    data : xarray.Dataset or numpy.ndarray
        ASCAT data.
    metadata : dict
        Metadata.
    """
    eps_file = read_eps(filename)
    ptype = eps_file.mphr["PRODUCT_TYPE"]
    fmv = int(eps_file.mphr["FORMAT_MAJOR_VERSION"])

    if ptype in ["SMR", "SMO"]:

        if fmv == 12:
            data, metadata = read_smx_fmv_12(eps_file)
        elif fmv == 11:
            data, metadata = read_smx_fmv_11(eps_file)
        else:
            raise RuntimeError("L2 SM format version not supported.")

        data["time"] = jd2dt(data.pop("jd"))

        rename_coords = {"longitude": "lon", "latitude": "lat"}

        for k, v in rename_coords.items():
            data[v] = data.pop(k)

        # convert spacecraft_id to internal sat_id
        sat_id = np.array([4, 3, 5])
        metadata["sat_id"] = sat_id[metadata["spacecraft_id"] - 1]

        # add/rename/remove fields according to generic format
        if generic:
            data = conv_epsl2szx_generic(data, metadata)

        # convert dict to xarray.Dataset or numpy.ndarray
        if to_xarray:
            for k in data.keys():
                if len(data[k].shape) == 1:
                    dim = ["obs"]
                elif len(data[k].shape) == 2:
                    dim = ["obs", "beam"]
                if k == "time":
                    data[k] = data[k].astype("datetime64[ns]")

                data[k] = (dim, data[k])

            coords = {}
            coords_fields = ["lon", "lat", "time"]
            for cf in coords_fields:
                coords[cf] = data.pop(cf)

            data = xr.Dataset(data, coords=coords, attrs=metadata)
            if generic:
                data = mask_dtype_nans(data)

        else:
            # collect dtype info
            dtype = []
            for var_name in data.keys():
                if len(data[var_name].shape) == 1:
                    dtype.append((var_name, data[var_name].dtype.str))
                elif len(data[var_name].shape) > 1:
                    dtype.append((var_name, data[var_name].dtype.str,
                                  data[var_name].shape[1:]))

            ds = np.empty(data["time"].size, dtype=np.dtype(dtype))
            for k, v in data.items():
                ds[k] = v
            data = ds
    else:
        raise ValueError("Format not supported. Product type {:1}"
                         " Format major version: {:2}".format(ptype, fmv))

    if return_ptype:
        metadata["product_type"] = ptype

    return data, metadata


def read_eps(filename,
             mphr_only=False,
             full=True,
             unsafe=False,
             scale_mdr=True):
    """
    Read EPS file.

    Parameters
    ----------
    filename : str
        Filename

    Returns
    -------
    prod : EPSProduct
        EPS data.
    """
    zipped = False
    if os.path.splitext(filename)[1] == ".gz":
        zipped = True

    # for zipped files use an unzipped temporary copy
    if zipped:
        with NamedTemporaryFile(delete=False) as tmp_fid:
            with GzipFile(filename) as gz_fid:
                tmp_fid.write(gz_fid.read())
            filename = tmp_fid.name

    # create the eps object with the filename and read it
    prod = EPSProduct(filename)

    if mphr_only:
        mphr = prod.read_mphr()
        prod.mphr = mphr
    else:
        prod.read(full, unsafe, scale_mdr)

    # remove the temporary copy
    if zipped:
        os.remove(filename)

    return prod


def read_szx_fmv_11(eps_file):
    """
    Read SZO/SZR format version 11.

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.

    Returns
    -------
    data : numpy.ndarray
        SZO/SZR data.
    """
    raw_data = eps_file.scaled_mdr
    raw_unscaled = eps_file.mdr
    mphr = eps_file.mphr

    n_node_per_line = raw_data["LONGITUDE"].shape[1]
    n_lines = raw_data["LONGITUDE"].shape[0]
    n_records = raw_data["LONGITUDE"].size

    data = {}
    metadata = {}
    idx_nodes = np.arange(n_lines).repeat(n_node_per_line)

    ascat_time = shortcdstime2jd(raw_data["UTC_LINE_NODES"].flatten()["day"],
                                 raw_data["UTC_LINE_NODES"].flatten()["time"])
    data["jd"] = ascat_time[idx_nodes]

    metadata["spacecraft_id"] = np.int8(mphr["SPACECRAFT_ID"][-1])
    metadata["orbit_start"] = np.uint32(mphr["ORBIT_START"])

    fields = [
        "processor_major_version", "processor_minor_version",
        "format_major_version", "format_minor_version"
    ]

    for f in fields:
        metadata[f] = np.int16(mphr[f.upper()])

    fields = ["sat_track_azi"]
    for f in fields:
        data[f] = raw_data[f.upper()].flatten()[idx_nodes]

    fields = [("longitude", long_nan), ("latitude", long_nan),
              ("swath_indicator", uint8_nan)]

    for f, nan_val in fields:
        data[f] = raw_data[f.upper()].flatten()
        valid = raw_unscaled[f.upper()].flatten() != nan_val
        data[f][~valid] = nan_val

    fields = [("sigma0_trip", long_nan), ("inc_angle_trip", uint_nan),
              ("azi_angle_trip", int_nan), ("kp", uint_nan),
              ("f_kp", uint8_nan), ("f_usable", uint8_nan), ("f_f", uint_nan),
              ("f_v", uint_nan), ("f_oa", uint_nan), ("f_sa", uint_nan),
              ("f_tel", uint_nan), ("f_land", uint_nan)]

    for f, nan_val in fields:
        data[f] = raw_data[f.upper()].reshape(n_records, 3)
        valid = raw_unscaled[f.upper()].reshape(n_records, 3) != nan_val
        data[f][~valid] = nan_val

    # modify longitudes from (0, 360) to (-180,180)
    mask = np.logical_and(data["longitude"] != long_nan, data["longitude"]
                          > 180)
    data["longitude"][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    mask = (data["azi_angle_trip"] != int_nan) & (data["azi_angle_trip"] < 0)
    data["azi_angle_trip"][mask] += 360

    data["node_num"] = np.tile((np.arange(n_node_per_line) + 1),
                               n_lines).astype(np.uint8)
    data["line_num"] = idx_nodes.astype(np.uint16)
    data["as_des_pass"] = (data["sat_track_azi"] < 270).astype(np.uint8)

    return data, metadata


def read_szx_fmv_12(eps_file):
    """
    Read SZO/SZR format version

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.

    Returns
    -------
    data : numpy.ndarray
        SZO/SZR data.
    """
    raw_data = eps_file.scaled_mdr
    raw_unscaled = eps_file.mdr
    mphr = eps_file.mphr

    n_node_per_line = raw_data["LONGITUDE"].shape[1]
    n_lines = raw_data["LONGITUDE"].shape[0]
    n_records = raw_data["LONGITUDE"].size

    data = {}
    metadata = {}
    idx_nodes = np.arange(n_lines).repeat(n_node_per_line)

    ascat_time = shortcdstime2jd(raw_data["UTC_LINE_NODES"].flatten()["day"],
                                 raw_data["UTC_LINE_NODES"].flatten()["time"])
    data["jd"] = ascat_time[idx_nodes]

    metadata["spacecraft_id"] = np.int8(mphr["SPACECRAFT_ID"][-1])
    metadata["orbit_start"] = np.uint32(mphr["ORBIT_START"])

    fields = [
        "processor_major_version", "processor_minor_version",
        "format_major_version", "format_minor_version"
    ]

    for f in fields:
        metadata[f] = np.int16(mphr[f.upper()])

    fields = [
        "degraded_inst_mdr", "degraded_proc_mdr", "sat_track_azi",
        "abs_line_number"
    ]

    for f in fields:
        data[f] = raw_data[f.upper()].flatten()[idx_nodes]

    fields = [("longitude", long_nan), ("latitude", long_nan),
              ("swath indicator", uint8_nan)]

    for f, nan_val in fields:
        data[f] = raw_data[f.upper()].flatten()
        valid = raw_unscaled[f.upper()].flatten() != nan_val
        data[f][~valid] = nan_val

    fields = [("sigma0_trip", long_nan), ("inc_angle_trip", uint_nan),
              ("azi_angle_trip", int_nan), ("kp", uint_nan),
              ("num_val_trip", ulong_nan), ("f_kp", uint8_nan),
              ("f_usable", uint8_nan), ("f_f", uint_nan), ("f_v", uint_nan),
              ("f_oa", uint_nan), ("f_sa", uint_nan), ("f_tel", uint_nan),
              ("f_ref", uint_nan), ("f_land", uint_nan)]

    for f, nan_val in fields:
        data[f] = raw_data[f.upper()].reshape(n_records, 3)
        valid = raw_unscaled[f.upper()].reshape(n_records, 3) != nan_val
        data[f][~valid] = nan_val

    # modify longitudes from (0, 360) to (-180,180)
    mask = np.logical_and(data["longitude"] != long_nan, data["longitude"]
                          > 180)
    data["longitude"][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    mask = (data["azi_angle_trip"] != int_nan) & (data["azi_angle_trip"] < 0)
    data["azi_angle_trip"][mask] += 360

    data["node_num"] = np.tile((np.arange(n_node_per_line) + 1),
                               n_lines).astype(np.uint8)

    data["line_num"] = idx_nodes.astype(np.uint16)

    data["as_des_pass"] = (data["sat_track_azi"] < 270).astype(np.uint8)

    data["swath_indicator"] = data.pop("swath indicator")

    return data, metadata


def read_szf_fmv_12(eps_file, ignore_noise_ool=False):
    """
    Read SZF format version 12.

    beam_num
    - 1 Left Fore Antenna
    - 2 Left Mid Antenna
    - 3 Left Aft Antenna
    - 4 Right Fore Antenna
    - 5 Right Mid Antenna
    - 6 Right Aft Antenna

    as_des_pass
    - 0 Ascending
    - 1 Descending

    swath_indicator
    - 0 Left
    - 1 Right

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.
    ignore_noise_ool : bool, optional
        Ignore noise out of limit flag (default: False).

    Returns
    -------
    data : numpy.ndarray
        SZF data.
    """
    data = {}
    metadata = {}

    n_lines = eps_file.mdr_counter
    n_node_per_line = eps_file.mdr["LONGITUDE_FULL"].shape[1]
    idx_nodes = np.arange(n_lines).repeat(n_node_per_line)

    # extract metadata
    metadata["spacecraft_id"] = np.int8(eps_file.mphr["SPACECRAFT_ID"][-1])
    metadata["orbit_start"] = np.uint32(eps_file.mphr["ORBIT_START"])
    metadata["state_vector_time"] = datetime.strptime(
        eps_file.mphr["STATE_VECTOR_TIME"][:-4], "%Y%m%d%H%M%S")

    fields = [
        "processor_major_version", "processor_minor_version",
        "format_major_version", "format_minor_version"
    ]
    for f in fields:
        metadata[f] = np.int16(eps_file.mphr[f.upper()])

    # extract time
    dt = np.datetime64(
        "2000-01-01") + eps_file.mdr["UTC_LOCALISATION"]["day"].astype(
            "timedelta64[D]"
        ) + eps_file.mdr["UTC_LOCALISATION"]["time"].astype("timedelta64[ms]")
    data["time"] = dt[idx_nodes]

    fields = [
        "degraded_inst_mdr", "degraded_proc_mdr", "sat_track_azi",
        "beam_number", "flagfield_rf1", "flagfield_rf2", "flagfield_pl",
        "flagfield_gen1"
    ]

    # 101 min = 6082 seconds
    # state_vector_time = ascending node crossing time - 1520.5,
    # time crossing at -90 lat
    orbit_start_time = metadata["state_vector_time"] - timedelta(
        seconds=1520.5)
    orbit_end_time = orbit_start_time + timedelta(seconds=6082)

    data["orbit_nr"] = np.ma.zeros(
        data["time"].size, dtype=np.int32,
        fill_value=int32_nan) + metadata["orbit_start"]
    data["orbit_nr"][data["time"] > orbit_end_time] += 1

    metadata["orbits"] = {}
    for orbit_nr in np.unique(data["orbit_nr"]):
        if orbit_nr == metadata["orbit_start"]:
            metadata["orbits"][orbit_nr] = (orbit_start_time, orbit_end_time)
        else:
            metadata["orbits"][orbit_nr] = (orbit_end_time, orbit_end_time +
                                            timedelta(seconds=6082))

    # extract data
    for f in fields:
        if eps_file.mdr_sfactor[f.upper()] == 1:
            data[f] = eps_file.mdr[f.upper()].flatten()[idx_nodes]
        else:
            data[f] = (eps_file.mdr[f.upper()].flatten() * 1. /
                       eps_file.mdr_sfactor[f.upper()])[idx_nodes]

    data["swath_indicator"] = (data["beam_number"].flatten()
                               > 3).astype(np.uint8)
    data["as_des_pass"] = (data["sat_track_azi"] < 270).astype(np.uint8)

    fields = [("longitude_full", long_nan), ("latitude_full", long_nan),
              ("sigma0_full", long_nan), ("inc_angle_full", uint_nan),
              ("azi_angle_full", int_nan), ("land_frac", uint_nan),
              ("flagfield_gen2", uint8_nan)]

    for f, nan_val in fields:
        data[f] = eps_file.mdr[f.upper()].flatten()
        invalid = eps_file.mdr[f.upper()].flatten() == nan_val

        if eps_file.mdr_sfactor[f.upper()] != 1:
            data[f] = data[f] * 1. / eps_file.mdr_sfactor[f.upper()]

        data[f][invalid] = nan_val

    # modify longitudes from (0, 360) to (-180, 180)
    mask = np.logical_and(data["longitude_full"] != long_nan,
                          data["longitude_full"] > 180)
    data["longitude_full"][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    idx = (data["azi_angle_full"] != int_nan) & (data["azi_angle_full"] < 0)
    data["azi_angle_full"][idx] += 360

    # set flags
    data["f_usable"] = set_flags(data, ignore_noise_ool)

    # create flagflield
    data["flagfield"] = gen_flagfield(data)

    return data, metadata


def read_smx_fmv_11(eps_file):
    """
    Read SMO/SMR format version 11.

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.

    Returns
    -------
    data : numpy.ndarray
        SMO/SMR data.
    """
    raw_data = eps_file.scaled_mdr
    raw_unscaled = eps_file.mdr

    n_node_per_line = raw_data["LONGITUDE"].shape[1]
    n_lines = raw_data["LONGITUDE"].shape[0]
    n_records = eps_file.mdr_counter * n_node_per_line
    idx_nodes = np.arange(eps_file.mdr_counter).repeat(n_node_per_line)

    data = {}
    metadata = {}

    metadata["spacecraft_id"] = np.int8(eps_file.mphr["SPACECRAFT_ID"][-1])
    metadata["orbit_start"] = np.uint32(eps_file.mphr["ORBIT_START"])

    ascat_time = shortcdstime2jd(raw_data["UTC_LINE_NODES"].flatten()["day"],
                                 raw_data["UTC_LINE_NODES"].flatten()["time"])
    data["jd"] = ascat_time[idx_nodes]

    fields = [("sigma0_trip", long_nan, long_nan), ("inc_angle_trip", uint_nan, uint_nan),
              ("azi_angle_trip", int_nan, int_nan), ("kp", uint_nan, uint_nan),
              ("f_land", uint_nan, float32_nan)]

    for f, nan_val, new_nan_val in fields:
        data[f] = raw_data[f.upper()].reshape(n_records, 3)
        valid = raw_unscaled[f.upper()].reshape(n_records, 3) != nan_val
        data[f][~valid] = new_nan_val

    fields = ["sat_track_azi"]
    for f in fields:
        data[f] = raw_data[f.upper()].flatten()[idx_nodes]

    fields = [("longitude", long_nan, long_nan),
              ("latitude", long_nan, long_nan),
              ("swath_indicator", uint8_nan, uint8_nan),
              ("soil_moisture", uint_nan, uint_nan),
              ("soil_moisture_error", uint_nan, uint_nan),
              ("sigma40", long_nan, long_nan),
              ("sigma40_error", long_nan, long_nan),
              ("slope40", long_nan, long_nan),
              ("slope40_error", long_nan, long_nan),
              ("dry_backscatter", long_nan, long_nan),
              ("wet_backscatter", long_nan, long_nan),
              ("mean_surf_soil_moisture", uint_nan, uint_nan),
              ("soil_moisture_sensetivity", ulong_nan, float32_nan),
              ("correction_flags", uint8_nan, uint8_nan),
              ("processing_flags", uint8_nan, uint8_nan),
              ("aggregated_quality_flag", uint8_nan, uint8_nan),
              ("snow_cover_probability", uint8_nan, uint8_nan),
              ("frozen_soil_probability", uint8_nan, uint8_nan),
              ("innudation_or_wetland", uint8_nan, uint8_nan),
              ("topographical_complexity", uint8_nan, uint8_nan)]

    for f, nan_val, new_nan_val in fields:
        data[f] = raw_data[f.upper()].flatten()
        valid = raw_unscaled[f.upper()].flatten() != nan_val
        data[f][~valid] = new_nan_val

    # sat_track_azi (uint)
    data["as_des_pass"] = \
        np.array(raw_data["SAT_TRACK_AZI"].flatten()[idx_nodes] < 270)

    # modify longitudes from [0,360] to [-180,180]
    mask = np.logical_and(data["longitude"] != long_nan, data["longitude"]
                          > 180)
    data["longitude"][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    mask = (data["azi_angle_trip"] != int_nan) & (data["azi_angle_trip"] < 0)
    data["azi_angle_trip"][mask] += 360

    fields = ["param_db_version", "warp_nrt_version"]
    for f in fields:
        data[f] = raw_data["PARAM_DB_VERSION"].flatten()[idx_nodes]

    metadata["spacecraft_id"] = int(eps_file.mphr["SPACECRAFT_ID"][2])

    data["node_num"] = np.tile((np.arange(n_node_per_line) + 1), n_lines)

    data["line_num"] = idx_nodes

    return data, metadata


def read_smx_fmv_12(eps_file):
    """
    Read SMO/SMR format version 12.

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.

    Returns
    -------
    data : numpy.ndarray
        SMO/SMR data.
    """
    raw_data = eps_file.scaled_mdr
    raw_unscaled = eps_file.mdr

    n_node_per_line = raw_data["LONGITUDE"].shape[1]
    n_lines = raw_data["LONGITUDE"].shape[0]
    n_records = eps_file.mdr_counter * n_node_per_line
    idx_nodes = np.arange(eps_file.mdr_counter).repeat(n_node_per_line)

    data = {}
    metadata = {}

    metadata["spacecraft_id"] = np.int8(eps_file.mphr["SPACECRAFT_ID"][-1])
    metadata["orbit_start"] = np.uint32(eps_file.mphr["ORBIT_START"])

    ascat_time = shortcdstime2jd(raw_data["UTC_LINE_NODES"].flatten()["day"],
                                 raw_data["UTC_LINE_NODES"].flatten()["time"])
    data["jd"] = ascat_time[idx_nodes]

    fields = [("sigma0_trip", long_nan, long_nan), ("inc_angle_trip", uint_nan, uint_nan),
              ("azi_angle_trip", int_nan, int_nan), ("kp", uint_nan, uint_nan),
              ("f_land", uint_nan, float32_nan)]

    for f, nan_val, new_nan_val in fields:
        data[f] = raw_data[f.upper()].reshape(n_records, 3)
        valid = raw_unscaled[f.upper()].reshape(n_records, 3) != nan_val
        data[f][~valid] = new_nan_val

    fields = ["sat_track_azi", "abs_line_number"]
    for f in fields:
        data[f] = raw_data[f.upper()].flatten()[idx_nodes]

    fields = [("longitude", long_nan, long_nan),
              ("latitude", long_nan, long_nan),
              ("swath_indicator", uint8_nan, uint8_nan),
              ("soil_moisture", uint_nan, uint_nan),
              ("soil_moisture_error", uint_nan, uint_nan),
              ("sigma40", long_nan, long_nan),
              ("sigma40_error", long_nan, long_nan),
              ("slope40", long_nan, long_nan),
              ("slope40_error", long_nan, long_nan),
              ("dry_backscatter", long_nan, long_nan),
              ("wet_backscatter", long_nan, long_nan),
              ("mean_surf_soil_moisture", uint_nan, uint_nan),
              ("soil_moisture_sensetivity", ulong_nan, float32_nan),
              ("correction_flags", uint8_nan, uint8_nan),
              ("processing_flags", uint8_nan, uint8_nan),
              ("aggregated_quality_flag", uint8_nan, uint8_nan),
              ("snow_cover_probability", uint8_nan, uint8_nan),
              ("frozen_soil_probability", uint8_nan, uint8_nan),
              ("innudation_or_wetland", uint8_nan, uint8_nan),
              ("topographical_complexity", uint8_nan, uint8_nan)]

    for f, nan_val, new_nan_val in fields:
        data[f] = raw_data[f.upper()].flatten()
        valid = raw_unscaled[f.upper()].flatten() != nan_val
        data[f][~valid] = new_nan_val

    # sat_track_azi (uint)
    data["as_des_pass"] = \
        np.array(raw_data["SAT_TRACK_AZI"].flatten()[idx_nodes] < 270)

    # modify longitudes from [0,360] to [-180,180]
    mask = np.logical_and(data["longitude"] != long_nan, data["longitude"]
                          > 180)
    data["longitude"][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    mask = (data["azi_angle_trip"] != int_nan) & (data["azi_angle_trip"] < 0)
    data["azi_angle_trip"][mask] += 360

    fields = ["param_db_version", "warp_nrt_version"]
    for f in fields:
        data[f] = raw_data["PARAM_DB_VERSION"].flatten()[idx_nodes]

    metadata["spacecraft_id"] = int(eps_file.mphr["SPACECRAFT_ID"][2])

    data["node_num"] = np.tile((np.arange(n_node_per_line) + 1), n_lines)

    data["line_num"] = idx_nodes

    return data, metadata


def read_szf_fmv_13(eps_file, ignore_noise_ool=False):
    """
    Read SZF format version 13.

    beam_num
    - 1 Left Fore Antenna
    - 2 Left Mid Antenna
    - 3 Left Aft Antenna
    - 4 Right Fore Antenna
    - 5 Right Mid Antenna
    - 6 Right Aft Antenna

    as_des_pass
    - 0 Ascending
    - 1 Descending

    swath_indicator
    - 0 Left
    - 1 Right

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.
    ignore_noise_ool : bool, optional
        Ignore noise out of limit flag (default: False).

    Returns
    -------
    data : numpy.ndarray
        SZF data.
    """
    data = {}
    metadata = {}

    n_lines = eps_file.mdr_counter
    n_node_per_line = eps_file.mdr["LONGITUDE_FULL"].shape[1]
    idx_nodes = np.arange(n_lines).repeat(n_node_per_line)

    # extract metadata
    metadata["spacecraft_id"] = np.int8(eps_file.mphr["SPACECRAFT_ID"][-1])
    metadata["orbit_start"] = np.uint32(eps_file.mphr["ORBIT_START"])
    metadata["state_vector_time"] = datetime.strptime(
        eps_file.mphr["STATE_VECTOR_TIME"][:-4], "%Y%m%d%H%M%S")

    fields = [
        "processor_major_version", "processor_minor_version",
        "format_major_version", "format_minor_version"
    ]
    for f in fields:
        metadata[f] = np.int16(eps_file.mphr[f.upper()])

    # extract time
    dt = np.datetime64(
        "2000-01-01") + eps_file.mdr["UTC_LOCALISATION"]["day"].astype(
            "timedelta64[D]"
        ) + eps_file.mdr["UTC_LOCALISATION"]["time"].astype("timedelta64[ms]")
    data["time"] = dt[idx_nodes]

    fields = [
        "degraded_inst_mdr", "degraded_proc_mdr", "sat_track_azi",
        "beam_number", "flagfield_rf1", "flagfield_rf2", "flagfield_pl",
        "flagfield_gen1"
    ]

    fields = [
        "degraded_inst_mdr", "degraded_proc_mdr", "sat_track_azi",
        "beam_number"
    ]

    # 101 min = 6082 seconds
    # state_vector_time = ascending node crossing time - 1520.5,
    # time crossing at -90 lat
    orbit_start_time = metadata["state_vector_time"] - timedelta(
        seconds=1520.5)
    orbit_end_time = orbit_start_time + timedelta(seconds=6082)

    data["orbit_nr"] = np.ma.zeros(
        data["time"].size, dtype=np.int32,
        fill_value=int32_nan) + metadata["orbit_start"]
    data["orbit_nr"][data["time"] > orbit_end_time] += 1

    metadata["orbits"] = {}
    for orbit_nr in np.unique(data["orbit_nr"]):
        if orbit_nr == metadata["orbit_start"]:
            metadata["orbits"][orbit_nr] = (orbit_start_time, orbit_end_time)
        else:
            metadata["orbits"][orbit_nr] = (orbit_end_time, orbit_end_time +
                                            timedelta(seconds=6082))

    # extract data
    for f in fields:
        if eps_file.mdr_sfactor[f.upper()] == 1:
            data[f] = eps_file.mdr[f.upper()].flatten()[idx_nodes]
        else:
            data[f] = (eps_file.mdr[f.upper()].flatten() * 1. /
                       eps_file.mdr_sfactor[f.upper()])[idx_nodes]

    data["swath_indicator"] = (data["beam_number"].flatten()
                               > 3).astype(np.uint8)
    data["as_des_pass"] = (data["sat_track_azi"] < 270).astype(np.uint8)

    fields = [("longitude_full", long_nan), ("latitude_full", long_nan),
              ("sigma0_full", long_nan), ("inc_angle_full", uint_nan),
              ("azi_angle_full", int_nan), ("flagfield", uint_nan)]

    for f, nan_val in fields:
        data[f] = eps_file.mdr[f.upper()].flatten()
        invalid = eps_file.mdr[f.upper()].flatten() == nan_val

        if eps_file.mdr_sfactor[f.upper()] != 1:
            data[f] = data[f] * 1. / eps_file.mdr_sfactor[f.upper()]

        data[f][invalid] = nan_val

    # modify longitudes from (0, 360) to (-180, 180)
    mask = np.logical_and(data["longitude_full"] != long_nan,
                          data["longitude_full"] > 180)
    data["longitude_full"][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    idx = (data["azi_angle_full"] != int_nan) & (data["azi_angle_full"] < 0)
    data["azi_angle_full"][idx] += 360

    # set flags
    data["f_usable"] = set_flags_fmv13(data["flagfield"], ignore_noise_ool)

    return data, metadata


def read_szx_fmv_13(eps_file):
    """
    Read SZO/SZR format version

    Parameters
    ----------
    eps_file : EPSProduct object
        EPS Product object.

    Returns
    -------
    data : numpy.ndarray
        SZO/SZR data.
    """
    raw_data = eps_file.scaled_mdr
    raw_unscaled = eps_file.mdr
    mphr = eps_file.mphr

    n_node_per_line = raw_data["LONGITUDE"].shape[1]
    n_lines = raw_data["LONGITUDE"].shape[0]
    n_records = raw_data["LONGITUDE"].size

    data = {}
    metadata = {}
    idx_nodes = np.arange(n_lines).repeat(n_node_per_line)

    ascat_time = shortcdstime2jd(raw_data["UTC_LINE_NODES"].flatten()["day"],
                                 raw_data["UTC_LINE_NODES"].flatten()["time"])
    data["jd"] = ascat_time[idx_nodes]

    metadata["spacecraft_id"] = np.int8(mphr["SPACECRAFT_ID"][-1])
    metadata["orbit_start"] = np.uint32(mphr["ORBIT_START"])

    fields = [
        "processor_major_version", "processor_minor_version",
        "format_major_version", "format_minor_version"
    ]

    for f in fields:
        metadata[f] = np.int16(mphr[f.upper()])

    fields = [
        "degraded_inst_mdr", "degraded_proc_mdr", "sat_track_azi",
        "abs_line_number"
    ]

    for f in fields:
        data[f] = raw_data[f.upper()].flatten()[idx_nodes]

    fields = [("longitude", long_nan), ("latitude", long_nan),
              ("swath indicator", int8_nan)]

    for f, nan_val in fields:
        data[f] = raw_data[f.upper()].flatten()
        valid = raw_unscaled[f.upper()].flatten() != nan_val
        data[f][~valid] = nan_val

    fields = [("sigma0_trip", long_nan), ("inc_angle_trip", uint_nan),
              ("azi_angle_trip", int_nan), ("kp", uint_nan),
              ("num_val_trip", ulong_nan), ("f_kp", uint8_nan),
              ("f_usable", int8_nan), ("land_frac", uint_nan)]

    for f, nan_val in fields:
        data[f] = raw_data[f.upper()].reshape(n_records, 3)
        valid = raw_unscaled[f.upper()].reshape(n_records, 3) != nan_val
        data[f][~valid] = nan_val

    # modify longitudes from (0, 360) to (-180,180)
    mask = np.logical_and(data["longitude"] != long_nan, data["longitude"]
                          > 180)
    data["longitude"][mask] += -360.

    # modify azimuth from (-180, 180) to (0, 360)
    mask = (data["azi_angle_trip"] != int_nan) & (data["azi_angle_trip"] < 0)
    data["azi_angle_trip"][mask] += 360

    data["node_num"] = np.tile((np.arange(n_node_per_line) + 1),
                               n_lines).astype(np.uint8)

    data["line_num"] = idx_nodes.astype(np.uint16)

    data["as_des_pass"] = (data["sat_track_azi"] < 270).astype(np.uint8)

    data["swath_indicator"] = data.pop("swath indicator")

    data["f_land"] = data.pop("land_frac")

    return data, metadata


def shortcdstime2jd(days, milliseconds):
    """
    Convert cds time to julian date.

    Parameters
    ----------
    days : int
        Days since 2000-01-01
    milliseconds : int
        Milliseconds.

    Returns
    -------
    jd : float
        Julian date.
    """
    offset = days + (milliseconds / 1000.) / (24. * 60. * 60.)
    return julian_epoch + offset


def set_flags(data, ignore_noise_ool=False):
    """
    Compute summary flag for each measurement with a value of 0, 1 or 2
    indicating nominal, slightly degraded or severely degraded data.

    The format of ASCAT products is defined by
    "EPS programme generic product format specification" (EPS.GGS.SPE.96167)
    and "ASCAT level 1 product format specification" (EPS.MIS.SPE.97233).

    bit name      category   description
    ------------------------------------

    flagfield_rf1
    0  fnoise     amber     noise missing, interpolated noise value used instead
    1  fpgp       amber     degraded power gain product
    2  vpgp       red       very degraded power gain product
    3  fhrx       amber     degraded filter shape
    4  vhrx       red       very degraded filter shape

    flagfield_rf2
    0  pgp_ool    red       power gain product is outside limits
    1  noise_ool  red       measured noise value is outside limits

    flagfield_pl
    0  forb       red       orbit height is outside limits
    1  fatt       red       no yaw steering
    2  fcfg       red       unexpected instrument configuration
    3  fman       red       satellite maneuver
    4  fosv       warning   osv file missing (fman may be incorrect)

    flagfield_gen1
    0  ftel       warning   telemetry missing (ftool may be incorrect)
    1  ftool      red       telemetry out of limits

    flagfield_gen2
    0  fsol   amber     possible interference from solar array
    1  fland  warning   lat/long position is over land
    2  fgeo   red       geolocation algorithm failed

    Each flag has belongs to a particular category which indicates the impact
    on data quality. Flags in the "amber" category indicate that the data is
    slightly degraded but still usable. Flags in the "red" category indicate
    that the data is severely degraded and should be discarded or
    used with caution.

    A simple algorithm for calculating a single summary flag with a value of
    0, 1 or 2 indicating nominal, slightly degraded or severely degraded is

    function calc_status( flags )
        status = 0
        if any amber flags are set then status = 1
        if any red flags are set then status = 2
    return status

    Parameters
    ----------
    data : numpy.ndarray
        SZF data.

    Returns
    -------
    f_usable : numpy.ndarray
        Flag indicating nominal (0), slightly degraded (1) or
        severely degraded(2).
    """
    flag_status_bit = {
        "flagfield_rf1": np.array([1, 1, 2, 1, 2, 0, 0, 0]),
        "flagfield_rf2": np.array([2, 2, 0, 0, 0, 0, 0, 0]),
        "flagfield_pl": np.array([2, 2, 2, 2, 0, 0, 0, 0]),
        "flagfield_gen1": np.array([0, 2, 0, 0, 0, 0, 0, 0]),
        "flagfield_gen2": np.array([1, 0, 2, 0, 0, 0, 0, 0])
    }

    if ignore_noise_ool:
        # remove "noise out of limits" as red flag
        flag_status_bit["flagfield_rf2"] = np.array([2, 0, 0, 0, 0, 0, 0, 0])

    f_usable = np.zeros(data["flagfield_rf1"].size, dtype=np.uint8)

    for flagfield, bitmask in flag_status_bit.items():
        subset = np.nonzero(data[flagfield])[0]

        if subset.size > 0:
            unpacked_bits = np.fliplr(
                np.unpackbits(data[flagfield][subset]).reshape(-1,
                                                               8).astype(bool))

            flag = np.ma.array(
                np.tile(bitmask, unpacked_bits.shape[0]).reshape(-1, 8),
                mask=~unpacked_bits,
                fill_value=0)

            f_usable[subset] = np.max(
                np.vstack((f_usable[subset], flag.filled().max(axis=1))),
                axis=0)

    return f_usable


def gen_flagfield(data):
    """
    The new flagfield collects the fields previously split across the RF1 /
    RF2 / PL / GEN1 / GEN2 flagfields. Its structure is described in the PFS,
    Tab. 14: Structure of FLAGFIELD.

    The old RF1 flagfield (related to the quality of the raw echo correction
    functions) contains the following bit flags and maps to the v11 flagfield
    as follows :

    RF1 Bit  Flag      v11 Bit  Description
    0        F_NOISE   0        Noise measurement missing, interpolated value used
    1        F_PG      1        Degraded power gain product
    2        V_PG      2        Very degraded power gain product
    3        F_FILTER  3        Degraded filter shape
    4        V_FILTER  4        Very degraded filter shape

    RF2 Bit  Flag         v11 Bit  Description
    0        F_PGP        5        Estimated power gain product outside limits
    1        F_NP         6        Measured noise outside limits
    2        F_PGP_DROP   7        Small drop in power gain product detected

    PL Bit   Flag          v11 Bit  Description
    0        F_ORBIT       n/a      Orbit height used for the NRCS normalisation is outside limits
    1        F_ATTITUDE    8        No yaw steering
    2        F_OMEGA       9        Unexpected instrument configuration
    3        F_MAN         10       Satellite manoeuvre
    4        F_OSV         11       Input orbit prediction file missing, OSV taken from L0 header

    GEN1 Bit  Flag           v11 Bit  Description
    0         F_E_TEL_PRES   12       Instrument or platform HKTM missing
    1         F_E_TEL_IR     13       Instrument or platform HKTM out of limits
    2         F_CE           n/a
    3         V_CE           n/a
    4         F_OA           n/a      Quality of satellite orbit and attitute
    5         F_TEL          n/a
    6         F_REF          14

    GEN2 Bit  Flag     v11 Bit  Description
    0         F_S_A    15       Potential interference from solar array
    1         F_LAND   16       Measurement over land in the generation of NCRS value
    2         F_GEO    17       Geolocation algorithm failed
    3         F_SIGN   18       The NRCS value is negative
    """
    flag_table = {
        "rf1": {
            0: 0,
            1: 1,
            2: 2,
            3: 3,
            4: 4
        },
        "rf2": {
            0: 5,
            1: 6,
            2: 7
        },
        "pl": {
            1: 8,
            2: 9,
            3: 10,
            4: 11
        },
        "gen1": {
            0: 12,
            1: 13,
            6: 14
        },
        "gen2": {
            0: 15,
            1: 16,
            2: 17,
            3: 18
        }
    }

    flagfield = np.zeros(data["flagfield_rf1"].size, dtype=np.uint32)

    for flag, table in flag_table.items():
        for sbit, tbit in table.items():
            pos = np.nonzero(get_bit(data[f"flagfield_{flag}"], sbit+1))[0]
            flagfield[pos] = set_bit(flagfield[pos], tbit+1)

    return flagfield


def set_flags_fmv13(flagfield, ignore_noise_ool=False):
    """
    Compute summary flag for each measurement with a value of 0, 1 or 2
    indicating nominal, slightly degraded or severely degraded data.

    The format of ASCAT products is defined by
    "EPS programme generic product format specification" (EPS.GGS.SPE.96167)
    and "ASCAT level 1 product format specification" (EPS.MIS.SPE.97233).

    bit name         category  description
    ------------------------------------
     0  f_noise       amber     1: noise missing/interpolated during processing
     1  f_pg          amber     1: degraded power gain product (pgp)
     2  v_pg          red       1: not valid power gain product (pgp)
     3  f_filter      amber     1: degraded hrx
     4  v_filter      red       1: no valid hrx
     5  f_pgp_ool     red       1: estimated power gain product out of limits
     6  f_np_ool      red       1: measured noise value is outside limits
     7  f_pgp_drop    amber     0: continuous pgp 1: drop in pgp
     8  f_attitude    red       1: non-normal attitude
     9  f_omega       red       1: instrument parameter configuration mismatch
    10  f_man         red       0: no-manoeuvre 1: manoeuvre
    11  f_osv         info      1: osv file not available
    12  f_e_tel_pres  amber     1: interpolated HKTM telemetry missing
    13  f_e_tel_ir    red       1: some interpolated HKTM telemetry parameters
                                   out of prescribed thresholds
    14  f_ref         info      1: if f_pgp or f_np are 1
    15  f_sa          amber     1: risk of solar array panel reflections
                                   interference
    16  f_land        info      0: no-land 1: land
    17  f_geo         red       1: geolocation algorithm failed
    18  f_sign        info         sigma0 in linear units is negative and value
                                   in dB has been calculated from its
                                   unsigned value
    19  f_com_op      info      1: data taken during commissioning phase

    20-31 spare

    Each flag has belongs to a particular category which indicates the impact
    on data quality. Flags in the "amber" category indicate that the data is
    slightly degraded but still usable. Flags in the "red" category indicate
    that the data is severely degraded and should be discarded or
    used with caution.

    Parameters
    ----------
    flagfield : numpy.ndarray
        Flags in decimal format.

    Returns
    -------
    f_usable : numpy.ndarray
        Flag indicating nominal (0), minor degraded (1) or major degraded (2).
    """
    # 0..ok, 1..minor/amber alert, 2..major/red alert
    bitmask = np.array(
        [1, 1, 2, 1, 2, 2, 2, 1, 2, 2, 2, 0, 1, 2, 0, 1, 0, 2, 0, 0],
        dtype=np.int8)

    if ignore_noise_ool:
        # remove "noise out of limits" as red flag
        bitmask[6] = 0

    # create look-up table
    def unpack(b):
        return np.clip(np.arange(2**bitmask.size) & 2**b, 0, 1) * bitmask[b]

    lut = np.max(list(map(unpack, list(range(bitmask.size)))), axis=0)
    f_usable = lut[flagfield].astype(np.int8)

    return f_usable



#!/usr/bin/env python3

import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import xarray as xr
import numpy as np

import ascat.read_native.generate_test_data as gtd

from ascat.grids import GridRegistry, NamedFileGridRegistry
from ascat.product_info import RaggedArrayCellProduct, OrthoMultiArrayCellProduct
from ascat.cell import RaggedArrayTs
from ascat.cell import OrthoMultiTimeseriesCell
from ascat.cell import CellGridFiles

from get_path import get_testdata_path

TESTDATA_PATH = get_testdata_path()

class RaggedArrayDummyCellProduct(RaggedArrayCellProduct):
    sample_dim = "time"
    grid_name = "fibgrid_12.5"

def add_sat_id(ds, sat_name):
    name_dict = {"metop_a": 3, "metop_b": 4, "metop_c": 5}
    if sat_name is not None:
        ds["sat_id"] = ("time", np.repeat(name_dict[sat_name], ds.sizes["time"]))
    return ds

def gen_dummy_cellfiles(dir, sat_name=None):
    contiguous_dir = dir / "contiguous"
    if sat_name is not None:
        contiguous_dir = contiguous_dir / sat_name
    contiguous_dir.mkdir(parents=True, exist_ok=True)
    indexed_dir = dir / "indexed"
    if sat_name is not None:
        indexed_dir = indexed_dir / sat_name
    indexed_dir.mkdir(parents=True, exist_ok=True)
    add_sat_id(gtd.contiguous_ragged_ds_2588, sat_name).to_netcdf(contiguous_dir / "2588.nc")
    add_sat_id(gtd.indexed_ragged_ds_2588, sat_name).to_netcdf(indexed_dir / "2588.nc")
    add_sat_id(gtd.contiguous_ragged_ds_2587, sat_name).to_netcdf(contiguous_dir / "2587.nc")
    add_sat_id(gtd.indexed_ragged_ds_2587, sat_name).to_netcdf(indexed_dir / "2587.nc")

class ERA5Cell(OrthoMultiArrayCellProduct):
    grid_name = "era5land"

    @classmethod
    def preprocessor(cls, ds):
        ds["location_id"].attrs["cf_role"] = "timeseries_id"
        return ds

class GLDASCell(OrthoMultiArrayCellProduct):
    grid_name = "gldas"

    @classmethod
    def preprocessor(cls, ds):
        ds["location_id"].attrs["cf_role"] = "timeseries_id"
        return ds

class TestOrthoMultiCellFile(unittest.TestCase):
    """
    Test the merge function
    """

    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.tempdir_path = Path(self.tempdir.name)
        gen_dummy_cellfiles(self.tempdir_path)

    def tearDown(self):
        self.tempdir.cleanup()

    def test_init(self):
        # contiguous_ragged_path = self.tempdir_path/ "contiguous" / "2588_contiguous_ragged.nc"
        gldas_path = TESTDATA_PATH / "warp/gldas_2023/"
        gldas_0029_path = TESTDATA_PATH / "warp/gldas_2023/0029.nc"
        gldas_0029 = OrthoMultiTimeseriesCell(gldas_0029_path)
        self.assertTrue(gldas_0029_path.samefile(gldas_0029.filenames[0]))

        cellnum_glob = "[0-9]" * 4 + ".nc"
        gldas_files = list(Path(gldas_path).glob(cellnum_glob))
        gldas = OrthoMultiTimeseriesCell(gldas_files)
        self.assertTrue(all([f in gldas.filenames
                             for f in gldas_files]))
        self.assertFalse("grid.nc" in [f.name for f in gldas.filenames])
        # self.assertIsNone(ra.ds))

    def test_read(self):
        gldas_path = TESTDATA_PATH / "warp/gldas_2023/"
        cellnum_glob = "[0-9]" * 4 + ".nc"
        gldas_files = list(Path(gldas_path).glob(cellnum_glob))
        gldas = OrthoMultiTimeseriesCell(gldas_files)
        # just make sure it works for now
        gldas.read()


    def test_cellgridfiles_read(self):
        gldas_path = TESTDATA_PATH / "warp/gldas_2023/"
        grid = gldas_path / "grid.nc"
        NamedFileGridRegistry.register("gldas", str(grid))
        del grid
        from time import time
        t1 = time()
        grid = GridRegistry().get("gldas")
        t2 = time()

        first_load_time = t2-t1

        for i in range(100):
            t1 = time()
            grid = GridRegistry().get("gldas")
            t2 = time()
            assert t2-t1 < first_load_time

        gldas_files = CellGridFiles.from_product_class(gldas_path, GLDASCell)
        ds = gldas_files.read()
        # assert it's a dataset and not empty
        assert isinstance(ds, xr.Dataset)
        assert len(ds) > 0

    def test_to_raster(self):
        gldas_path = TESTDATA_PATH / "warp/gldas_2023/"
        grid = gldas_path / "grid.nc"
        NamedFileGridRegistry.register("gldas", str(grid))
        del grid

        gldas_files = CellGridFiles.from_product_class(gldas_path, GLDASCell)
        gldas_files.read().cf_geom.to_raster(x_var="lon", y_var="lat")





class TestRaggedArrayCellFile(unittest.TestCase):
    """
    Test the merge function
    """

    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.tempdir_path = Path(self.tempdir.name)
        self.indexed_cells_path = TESTDATA_PATH / "hsaf/h129/stack_cells/"
        gen_dummy_cellfiles(self.tempdir_path)

    def tearDown(self):
        self.tempdir.cleanup()

    def test_init(self):
        contiguous_ragged_path = self.tempdir_path/ "contiguous" / "2588_contiguous_ragged.nc"
        ra = RaggedArrayTs(contiguous_ragged_path)
        self.assertEqual(ra.filenames[0], contiguous_ragged_path)
        # self.assertIsNone(ra.ds)

    def test_read(self):
        contiguous_ragged_path = self.tempdir_path / "contiguous" / "2588.nc"
        ra = RaggedArrayTs(contiguous_ragged_path)
        ds = ra.read(chunks={"obs": 3})
        self.assertIsInstance(ds, xr.Dataset)
        self.assertIn("lon", ds)
        self.assertIn("lat", ds)
        self.assertIn("row_size", ds)
        # self.assertIn("time", ds)
        self.assertIn("obs", ds.dims)
        # assert chunk size
        # self.assertEqual(ds["lon"].data.chunksize, (5,))
        #

    def test__ensure_obs(self):
        contiguous_ragged_path = self.tempdir_path / "contiguous" / "2588.nc"
        ra = RaggedArrayTs(contiguous_ragged_path)
        ds = xr.open_dataset(contiguous_ragged_path)
        # original_dim = ds["time"]
        self.assertNotIn("obs", ds.dims)
        self.assertIn("time", ds.dims)
        ds = ra._ensure_obs(ds)
        self.assertIn("obs", ds.dims)
        self.assertNotIn("time", ds.dims)
        # print(original_dim.values)
        # np.testing.assert_equal(ds["obs"].values, original_dim.values)

    def test__indexed_or_contiguous(self):
        contiguous_ragged_path = self.tempdir_path / "contiguous" / "2588.nc"
        ra = RaggedArrayTs(contiguous_ragged_path)
        ds = xr.open_dataset(contiguous_ragged_path)
        self.assertEqual(ra._indexed_or_contiguous(ds), "contiguous")

    def test__ensure_indexed(self):
        contiguous_ragged_path = self.tempdir_path / "contiguous" / "2588.nc"
        ra = RaggedArrayTs(contiguous_ragged_path)
        ds = xr.open_dataset(contiguous_ragged_path)
        self.assertNotIn("locationIndex", ds)
        self.assertIn("row_size", ds)

        self.assertIsNone(ra._ensure_indexed(None))

        indexed_ds = ra._ensure_indexed(ds)
        self.assertIn("locationIndex", indexed_ds)
        self.assertNotIn("row_size", indexed_ds)

        xr.testing.assert_equal(ra._ensure_indexed(indexed_ds), indexed_ds)

        np.testing.assert_array_equal(indexed_ds["locationIndex"].values,
                                      np.array([0, 0, 1, 2, 2, 3, 3, 3, 3, 4],
                                               dtype=np.int32))

    def test__ensure_contiguous(self):
        indexed_ragged_path = self.tempdir_path / "indexed" / "2588.nc"
        ra = RaggedArrayTs(indexed_ragged_path)
        ds = xr.open_dataset(indexed_ragged_path)
        ds = ds.chunk({"time": 1_000_000})
        self.assertIn("locationIndex", ds)
        self.assertNotIn("row_size", ds)

        self.assertIsNone(ra._ensure_contiguous(None))

        contiguous_ds = ra._ensure_contiguous(ds)
        self.assertNotIn("locationIndex", contiguous_ds)
        self.assertIn("row_size", contiguous_ds)

        xr.testing.assert_equal(ra._ensure_contiguous(contiguous_ds), contiguous_ds)

        np.testing.assert_array_equal(contiguous_ds["row_size"].values,
                                        np.array([2, 1, 2, 4, 1], dtype=np.int32))

    def test__trim_to_gpis(self):
        indexed_ragged_path = self.tempdir_path / "indexed" / "2588.nc"
        ra = RaggedArrayTs(indexed_ragged_path)
        ds = xr.open_dataset(indexed_ragged_path)
        ds = ra._ensure_obs(ds)
        ds = ds.chunk({"obs": 1_000_000})
        valid_gpis = [1549346, 1555912]
        trimmed_ds = ra._trim_to_gpis(ds, valid_gpis)
        self.assertTrue(np.all(np.isin(trimmed_ds["location_id"].values, valid_gpis)))
        new_obs_gpis = trimmed_ds["location_id"].values[trimmed_ds["locationIndex"].values]
        np.testing.assert_array_equal(new_obs_gpis, np.repeat(np.array([1549346, 1555912]), [2, 4]))

    def test__trim_var_range(self):
        indexed_ragged_path = self.tempdir_path / "indexed" / "2588.nc"
        ra = RaggedArrayTs(indexed_ragged_path)
        ds = xr.open_dataset(indexed_ragged_path)
        ds = ra._ensure_obs(ds)
        ds = ds.chunk({"obs": 1_000_000})
        start_dt = np.datetime64("2020-01-01T00:00:02")
        end_dt = np.datetime64("2020-01-01T00:00:04")
        trimmed_ds = ra._trim_var_range(ds, "time", start_dt, end_dt)
        self.assertTrue(np.all(trimmed_ds["time"].values < end_dt))
        self.assertTrue(np.all(trimmed_ds["time"].values >= start_dt))


    def test_back_and_forth(self):
        contiguous_ragged_path = self.tempdir_path / "contiguous" / "2587.nc"
        ra = RaggedArrayTs(contiguous_ragged_path)
        orig_ds = xr.open_dataset(contiguous_ragged_path)
        ds = ra._ensure_contiguous(orig_ds)
        xr.testing.assert_equal(orig_ds, ds)

    def test_merge(self):
        fname1 = self.tempdir_path / "contiguous" / "2588.nc"
        fname2 = self.tempdir_path / "contiguous" / "2587.nc"
        ra1 = RaggedArrayTs(fname1)
        ra2 = RaggedArrayTs(fname2)
        ds1 = ra1.read(return_format="point")
        ds2 = ra2.read(return_format="point")
        merged = ra1.merge([ds1, ds2]).cf_geom.to_indexed_ragged()
        np.testing.assert_array_equal(merged["locationIndex"].values,
                                        np.array([5, 5, 6, 7, 7, 8, 8, 8, 8, 9, 0, 0, 0, 1, 2, 3, 3, 3, 4, 4],
                                                 dtype=np.int32))
        contig = ra1._ensure_contiguous(merged)
        np.testing.assert_array_equal(contig["row_size"].values,
                                        np.array([3, 1, 1, 3, 2, 2, 1, 2, 4, 1], dtype=np.int32))

        self.assertIsNone(ra1.merge([]))

    def test__merge_contiguous(self):
        fname1 = self.tempdir_path / "contiguous" / "2588.nc"
        fname2 = self.tempdir_path / "contiguous" / "2587.nc"
        ra = RaggedArrayTs([fname1, fname2])
        ds = ra.read()


    # def test__trim_var_range(self):
    #     contiguous_ragged_fname = "2588_contiguous_ragged.nc"
    #     ra = RaggedArray(self.tempdir_path / contiguous_ragged_fname)
    #     ra.read()

    # def test__location_vars_from_ds_list(self):
    #     contiguous_ragged_fname = "2588_contiguous_ragged.nc"
    #     ra = RaggedArray(self.tempdir_path / contiguous_ragged_fname)
    #     ra.read()
    #     # print(ra._location_vars_from_ds_list([ra.ds]))

    # def test__only_locations(self):
    #     contiguous_ragged_fname = "2588_contiguous_ragged.nc"
    #     ra = RaggedArray(self.tempdir_path / contiguous_ragged_fname)
    #     ra.read()
    #     xr.testing.assert_equal(ra._only_locations(ra.ds),
    #                             ra.ds[["lon", "lat", "location_id"]])
        # print(ra._only_locations(ra.ds))


    # def test_read_period(self):
        # ra = RaggedArray(self.tempdir_path / "contiguous_ragged.nc")
        # ra.read_period(datetime(2020, 1, 1), datetime(2020, 1, 2))

grid_registry = GridRegistry()
class TestCellGridFiles(unittest.TestCase):
    """
    Test a collection of cell files
    """

    @staticmethod
    def _init_options(root_path, sf_templ=None, sf_read_fmt=None):
        return {
            "root_path": root_path,
            "file_class": RaggedArrayTs,
            "fn_format": "{:04d}.nc",
            #"sf_format": sf_templ,
            "grid": grid_registry.get("fibgrid_12.5"),
        }

    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.tempdir_path = Path(self.tempdir.name)
        # self.indexed_cells_path = TESTDATA_PATH / "hsaf/h129/stack_cells/"
        gen_dummy_cellfiles(self.tempdir_path)
        gen_dummy_cellfiles(self.tempdir_path, "metop_a")
        gen_dummy_cellfiles(self.tempdir_path, "metop_b")
        gen_dummy_cellfiles(self.tempdir_path, "metop_c")
        self.indexed_cells_path = self.tempdir_path / "indexed" / "metop_a"
        self.contiguous_cells_path = self.tempdir_path / "contiguous" / "metop_a"



    def tearDown(self):
        self.tempdir.cleanup()

    def test_init(self):
        contig_collection = CellGridFiles(
            **self._init_options(self.tempdir_path / "contiguous")
        )
        self.assertEqual(contig_collection._fn(2588).name, "2588.nc")
        self.assertEqual(contig_collection.root_path, self.tempdir_path / "contiguous")
        self.assertEqual(contig_collection.file_class, RaggedArrayTs)

    def test_spatial_search(self):
        root_path = self.tempdir_path / "contiguous" / "metop_a"
        contig_collection = CellGridFiles(
            **self._init_options(root_path)
        )
        self.assertEqual(contig_collection.spatial_search(),
                         [root_path/"2587.nc", root_path/"2588.nc"])
        self.assertEqual(contig_collection.spatial_search(location_id=1549346),
                         [root_path/"2588.nc"])
        self.assertEqual(contig_collection.spatial_search(location_id=1493629),
                         [root_path/"2587.nc"])
        self.assertEqual(contig_collection.spatial_search(location_id=[1549346, 1493629]),
                         [root_path/"2587.nc", root_path/"2588.nc"])
        self.assertEqual(contig_collection.spatial_search(location_id=[1493629, 1549346, 1493629]),
                         [root_path/"2587.nc", root_path/"2588.nc"])

    def test_read(self):
        root_path = self.tempdir_path / "contiguous"
        contig_collection = CellGridFiles(
            **self._init_options(root_path, {"sat_str": "{sat}"}, {"sat_str": {"sat": "metop_[abc]"}})
        )
        ds_2588 = contig_collection.read(cell=2588)
        self.assertIsInstance(ds_2588, xr.Dataset)

        ds_2587 = contig_collection.read(cell=2587)
        self.assertIsInstance(ds_2587, xr.Dataset)

        ds_cell_merged = contig_collection.read(cell=[2587, 2588])
        self.assertIsInstance(ds_cell_merged, xr.Dataset)


    def test_read_with_one_coord(self):
        root_path = self.tempdir_path / "contiguous"
        contig_collection = CellGridFiles(
            **self._init_options(root_path, {"sat_str": "{sat}"}, {"sat_str": {"sat": "metop_[abc]"}})
        )
        coord = (175.8, 70.01)
        ds = contig_collection.read(coords=coord)
        ds.load()
        assert len(ds.time) >= 1

    def test_read_with_two_valid_coords(self):
        root_path = self.tempdir_path / "contiguous"
        contig_collection = CellGridFiles(
            **self._init_options(root_path, {"sat_str": "{sat}"}, {"sat_str": {"sat": "metop_[abc]"}})
        )
        coords = (np.array([175.8, 175.4]),
                  np.array([70.01, 70.05]))
        ds = contig_collection.read(coords=coords)
        assert len(ds.time) >= 1

    def test_read_with_invalid_coords(self):
        root_path = self.tempdir_path / "contiguous"
        contig_collection = CellGridFiles(
            **self._init_options(root_path, {"sat_str": "{sat}"}, {"sat_str": {"sat": "metop_[abc]"}})
        )
        coords = (np.array([1, 2, 0]),
                  np.array([10, 20, 0]))
        ds = contig_collection.read(coords=coords)
        assert ds is None

    def test_read_with_two_valid_coords_and_invalid_coord(self):
        root_path = self.tempdir_path / "contiguous"
        contig_collection = CellGridFiles(
            **self._init_options(root_path, {"sat_str": "{sat}"}, {"sat_str": {"sat": "metop_[abc]"}})
        )
        coords = (np.array([175.8, 175.4, 0]),
                  np.array([70.01, 70.05, 0]))
        ds = contig_collection.read(coords=coords)
        ds.load()
        assert len(ds.time) >= 1

    def test_read_indexed(self):
        root_path = self.tempdir_path / "indexed"
        indexed_collection = CellGridFiles(
            **self._init_options(root_path, {"sat_str": "{sat}"}, {"sat_str": {"sat": "metop_[abc]"}})
        )
        ds = indexed_collection.read(bbox=(70.5, 75, 175.5, 179))

    def test_read_contiguous(self):
        root_path = self.tempdir_path / "contiguous"
        indexed_collection = CellGridFiles(
            **self._init_options(root_path, {"sat_str": "{sat}"}, {"sat_str": {"sat": "metop_[abc]"}})
        )
        ds = indexed_collection.read(bbox=(70.5, 75, 175.5, 179))
        assert len(ds.time) >= 1

    def test_read_single_ts_indexed(self):
        files = list(self.indexed_cells_path.glob("*.nc"))
        first_file_ds = xr.open_dataset(files[0])
        one_valid_gpi = [first_file_ds["location_id"][first_file_ds["locationIndex"][5]].values]
        collection = CellGridFiles.from_product_class(self.indexed_cells_path, RaggedArrayDummyCellProduct)
        ds = collection.read(location_id=one_valid_gpi)
        assert len(ds.cf_geom.to_indexed_ragged().locations) == 1
        assert len(ds.time) >= 1

    def test_read_n_ts_from_one_cell_indexed(self):
        n = 5
        files = list(self.indexed_cells_path.glob("*.nc"))
        first_file_ds = xr.open_dataset(files[0])

        n_valid_gpis = np.unique([first_file_ds["location_id"][first_file_ds["locationIndex"]].values])[:n]
        collection = CellGridFiles.from_product_class(self.indexed_cells_path, RaggedArrayDummyCellProduct)

        # Try reading all at once
        ds = collection.read(location_id=n_valid_gpis)
        idxed_ds = ds.cf_geom.to_indexed_ragged().load()
        assert len(idxed_ds.locations) == n

        # Try reading one by one in a loop
        # This access pattern is slow because under the hood we convert to point array on every read.
        # Better to read all the data at once and then select the points when doing indexed.
        for gpi in n_valid_gpis:
            collection.read(location_id=gpi, return_format="point")

    def test_read_2n_ts_from_two_cells_indexed(self):
        n = 5
        files = list(self.indexed_cells_path.glob("*.nc"))
        first_file_ds = xr.open_dataset(files[0])
        second_file_ds = xr.open_dataset(files[1])

        n_valid_gpis_1 = np.unique([first_file_ds["location_id"][first_file_ds["locationIndex"]].values])[:n]
        n_valid_gpis_2 = np.unique([second_file_ds["location_id"][second_file_ds["locationIndex"]].values])[:n]
        n_times_2_valid_gpis = np.concatenate([n_valid_gpis_1, n_valid_gpis_2])

        collection = CellGridFiles.from_product_class(self.indexed_cells_path, RaggedArrayDummyCellProduct)

        # Try reading all at once
        ds = collection.read(location_id=n_times_2_valid_gpis)
        idxed_ds = ds.cf_geom.to_indexed_ragged().load()
        assert len(idxed_ds.locations) == 2*n

    def test_read_one_cell_indexed(self):
        files = list(self.indexed_cells_path.glob("*.nc"))
        first_cell = int(files[0].stem)

        collection = CellGridFiles.from_product_class(self.indexed_cells_path, RaggedArrayDummyCellProduct)
        ds = collection.read(cell=first_cell)
        assert len(ds.time) >= 1

    def test_read_two_cells_indexed(self):
        files = list(self.indexed_cells_path.glob("*.nc"))
        first_cell = int(files[0].stem)
        second_cell = int(files[1].stem)

        collection = CellGridFiles.from_product_class(self.indexed_cells_path, RaggedArrayDummyCellProduct)
        ds = collection.read(cell=[first_cell, second_cell])
        assert len(ds.time) >= 1

    def test_read_single_ts_contiguous(self):
        files = list(self.contiguous_cells_path.glob("*.nc"))
        first_file_ds = xr.open_dataset(files[0])
        valid_gpis = np.unique(np.repeat(first_file_ds["location_id"].values, first_file_ds["row_size"].values))
        one_valid_gpi = [valid_gpis[0]]
        collection = CellGridFiles.from_product_class(self.contiguous_cells_path, RaggedArrayDummyCellProduct)
        ds = collection.read(location_id=one_valid_gpi)
        assert len(ds.time) >= 1

    def test_read_n_ts_from_one_cell_contiguous(self):
        n = 5
        files = list(self.contiguous_cells_path.glob("*.nc"))
        first_file_ds = xr.open_dataset(files[0])

        valid_gpis = np.unique(np.repeat(first_file_ds["location_id"].values, first_file_ds["row_size"].values))
        n_valid_gpis = valid_gpis[:n]
        collection = CellGridFiles.from_product_class(self.contiguous_cells_path, RaggedArrayDummyCellProduct)

        # Try reading all at once
        ds = collection.read(location_id=n_valid_gpis)
        idxed_ds = ds.cf_geom.to_contiguous_ragged().load()
        assert len(idxed_ds.locations) == n

        # Try reading one by one in a loop
        for gpi in n_valid_gpis:
            new_ds = collection.read(location_id=gpi, return_format="point")
            assert len(new_ds.time) >= 1

    def test_read_2n_ts_from_two_cells_contiguous(self):
        n = 5
        files = list(self.contiguous_cells_path.glob("*.nc"))
        first_file_ds = xr.open_dataset(files[0])
        second_file_ds = xr.open_dataset(files[1])

        valid_gpis_1 = np.unique(np.repeat(first_file_ds["location_id"].values, first_file_ds["row_size"].values))
        n_valid_gpis_1 = valid_gpis_1[:n]
        valid_gpis_2 = np.unique(np.repeat(second_file_ds["location_id"].values, second_file_ds["row_size"].values))
        n_valid_gpis_2 = valid_gpis_2[:n]
        n_times_2_valid_gpis = np.concatenate([n_valid_gpis_1, n_valid_gpis_2])

        collection = CellGridFiles.from_product_class(self.contiguous_cells_path, RaggedArrayDummyCellProduct)

        # Try reading all at once
        ds = collection.read(location_id=n_times_2_valid_gpis)
        idxed_ds = ds.cf_geom.to_contiguous_ragged().load()
        assert len(idxed_ds.locations) == 2*n

    def test_read_one_cell_contiguous(self):
        files = list(self.contiguous_cells_path.glob("*.nc"))
        first_cell = int(files[0].stem)

        collection = CellGridFiles.from_product_class(self.contiguous_cells_path, RaggedArrayDummyCellProduct)
        ds = collection.read(cell=first_cell)
        assert len(ds.time) >= 1

    def test_read_two_cells_contiguous(self):
        files = list(self.contiguous_cells_path.glob("*.nc"))
        first_cell = int(files[0].stem)
        second_cell = int(files[1].stem)

        collection = CellGridFiles.from_product_class(self.contiguous_cells_path, RaggedArrayDummyCellProduct)
        ds = collection.read(cell=[first_cell, second_cell])
        assert len(ds.time) >= 1

if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import warnings

import netCDF4
import numpy as np
import pygeogrids.grids as grids
from pynetcf.time_series import GriddedNcContiguousRaggedTs

float32_nan = -999999.0


class TimeSeries:

    """
    Container class for a time series.

    Parameters
    ----------
    gpi : int
        Grid point index
    lon : float
        Longitude of grid point
    lat : float
        Latitude of grid point
    cell : int
        Cell number of grid point
    data : pandas.DataFrame
        DataFrame which contains the data
    topo_complex : int, optional
        Topographic complexity at the grid point
    wetland_frac : int, optional
        Wetland fraction at the grid point
    porosity_gldas : float, optional
        Porosity taken from GLDAS model
    porosity_hwsd : float, optional
        Porosity calculated from Harmonised World Soil Database

    Attributes
    ----------
    gpi : int
        Grid point index
    lon : float
        Longitude of grid point
    lat : float
        Latitude of grid point
    cell : int
        Cell number of grid point
    data : pandas.DataFrame
        DataFrame which contains the data
    topo_complex : int
        Topographic complexity at the grid point
    wetland_frac : int
        Wetland fraction at the grid point
    porosity_gldas : float
        Porosity taken from GLDAS model
    porosity_hwsd : float
        Porosity calculated from Harmonised World Soil Database
    """

    def __init__(self, gpi, lon, lat, cell, data,
                 topo_complex=None, wetland_frac=None,
                 porosity_gldas=None, porosity_hwsd=None):

        self.gpi = gpi
        self.lon = lon
        self.lat = lat
        self.data = data
        self.cell = cell
        self.topo_complex = topo_complex
        self.wetland_frac = wetland_frac
        self.porosity_gldas = porosity_gldas
        self.porosity_hwsd = porosity_hwsd

    def __repr__(self):
        msg = "GPI: {:d} Lon: {:2.3f} Lat: {:3.3f}".format(
            self.gpi, self.lon, self.lat)

        return msg


def load_grid(filename):
    """
    Load grid file.

    Parameters
    ----------
    filename : str
        Grid filename.

    Returns
    -------
    grid : pygeogrids.CellGrid
        Grid.
    """
    with netCDF4.Dataset(filename) as grid_nc:
        land_gp = np.where(grid_nc.variables['land_flag'][:] == 1)[0]
        lons = grid_nc.variables['lon'][:]
        lats = grid_nc.variables['lat'][:]
        gpis = grid_nc.variables['gpi'][:]
        cells = grid_nc.variables['cell'][:]

    grid = grids.CellGrid(lons[land_gp], lats[land_gp], cells[land_gp],
                          gpis[land_gp])

    return grid


class StaticLayers():

    """
    Class to read static layer files.

    Parameters
    ----------
    path : str
        Path of static layer files.
    topo_wetland_file : str, optional
        Topographic and complexity file (default: None).
    frozen_snow_file : str, optional
        Frozen and snow cover probability file (default: None).
    porosity_file : str, optional
        Porosity file (default: None).
    cache : bool, optional
        If true all static layers are loaded into memory (default: False).

    Attributes
    ----------
    topo_wetland : dict
        Topographic complexity and inundation and wetland fraction.
    frozen_snow_prob : dict
        Frozen soil/canopy probability and snow cover probability.
    porosity : dict
        Soil porosity information.
    """

    def __init__(self, path, topo_wetland_file=None,
                 frozen_snow_file=None, porosity_file=None, cache=False):

        if cache:
            print("Static layers will be loaded, this may take some time.")

        if topo_wetland_file is None:
            topo_wetland_file = os.path.join(path, 'topo_wetland.nc')

        self.topo_wetland = StaticFile(topo_wetland_file,
                                       ['wetland', 'topo'],
                                       cache=cache)

        if frozen_snow_file is None:
            frozen_snow_file = os.path.join(path, 'frozen_snow_probability.nc')

        self.frozen_snow_prob = StaticFile(frozen_snow_file,
                                           ['snow_prob', 'frozen_prob'],
                                           cache=cache)
        if porosity_file is None:
            porosity_file = os.path.join(path, 'porosity.nc')

        self.porosity = StaticFile(porosity_file,
                                   ['por_gldas', 'por_hwsd'],
                                   cache=cache)


class StaticFile:

    """
    StaticFile class.

    Parameters
    ----------
    filename : str
        File name.
    variables : list of str
        List of variables.
    cache : bool, optional
        Flag to cache data stored in file (default: False).

    Attributes
    ----------
    filename : str
        Static layer file name.
    variables : list of str
        List of variables.
    cache : bool
        Flag to cache data stored in file.
    data : dict
        Dictionary containing static layer data.
    """

    def __init__(self, filename, variables, cache=False):
        self.filename = filename
        self.cache = cache
        self.variables = variables
        self.data = {}

        if self.cache:
            with netCDF4.Dataset(self.filename) as nc_file:
                for v in self.variables:
                    self.data[v] = nc_file.variables[v][:].filled()

    def __getitem__(self, gpi):
        """
        Get data at given GPI.

        Parameters
        ----------
        gpi : int
            Grid point index.
        """
        data = {}
        if self.cache:
            for v in self.variables:
                data[v] = self.data[v][gpi]
        else:
            with netCDF4.Dataset(self.filename) as nc_file:
                for v in self.variables:
                    data[v] = nc_file.variables[v][[gpi]].filled()[0]

        return data


class AscatGriddedNcTs(GriddedNcContiguousRaggedTs):

    """
    Class reading Metop ASCAT soil moisture Climate Data Record (CDR).

    Parameters
    ----------
    path : str
        Path to Climate Data Record (CDR) data set.
    fn_format : str
        Filename format string, typical '<prefix>_{:04d}'
    grid_filename : str
        Grid filename.
    static_layer_path : str, optional
        Path to static layer files (default: None).
    thresholds : dict, optional
        Thresholds for topographic complexity (default 50) and
        wetland fraction (default 50).

    Attributes
    ----------
    grid : pygeogrids.CellGrid
        Cell grid.
    thresholds : dict
        Thresholds for topographic complexity (default 50) and
        wetland fraction (default 50).
    slayer : str
        StaticLayer object
    """

    def __init__(self, path, fn_format, grid_filename, static_layer_path=None,
                 cache_static_layer=False, thresholds=None, **kwargs):

        grid = load_grid(grid_filename)

        self.thresholds = {'topo_complex': 50, 'wetland_frac': 50}

        if thresholds is not None:
            self.thresholds.update(thresholds)

        self.slayer = None

        if static_layer_path is not None:
            self.slayer = StaticLayers(static_layer_path,
                                       cache=cache_static_layer)

        super().__init__(path, grid, fn_format=fn_format,
                         **kwargs)

    def _read_gp(self, gpi, **kwargs):
        """
        Read time series for specific grid point.

        Parameters
        ----------
        gpi : int
            Grid point index.
        mask_ssf : boolean, optional
            Default False, if True only SSF values of 1 and 0 will be
            allowed, all others are removed
        mask_frozen_prob : int, optional
            If included in kwargs then all observations taken when
            frozen probability > mask_frozen_prob are removed from the data
            Default: no masking
        mask_snow_prob : int, optional
            If included in kwargs then all observations taken when
            snow probability > mask_snow_prob are removed from the data

        Returns
        -------
        ts : AscatTimeSeries
            Time series object.
        """
        absolute_sm = kwargs.pop('absolute_sm', None)
        mask_frozen_prob = kwargs.pop('mask_frozen_prob', None)
        mask_snow_prob = kwargs.pop('mask_snow_prob', None)
        mask_ssf = kwargs.pop('mask_ssf', None)

        data = super()._read_gp(gpi, **kwargs)
        data.attrs = {}
        data.attrs['gpi'] = gpi
        data.attrs['lon'], data.attrs['lat'] = self.grid.gpi2lonlat(gpi)
        data.attrs['cell'] = self.grid.gpi2cell(gpi)

        if self.slayer is not None:
            data.attrs['topo_complex'] = self.slayer.topo_wetland[gpi]['topo']
            data.attrs['wetland_frac'] = self.slayer.topo_wetland[gpi]['wetland']
            snow_prob = self.slayer.frozen_snow_prob[gpi]['snow_prob']
            frozen_prob = self.slayer.frozen_snow_prob[gpi]['frozen_prob']
            data.attrs['porosity_gldas'] = self.slayer.porosity[gpi]['por_gldas']
            data.attrs['porosity_hwsd'] = self.slayer.porosity[gpi]['por_hwsd']

            if data.attrs['porosity_gldas'] == float32_nan:
                data.attrs['porosity_gldas'] = np.nan

            if data.attrs['porosity_hwsd'] == float32_nan:
                data.attrs['porosity_hwsd'] = np.nan

            if data is not None:
                data['snow_prob'] = snow_prob[data.index.dayofyear - 1]
                data['frozen_prob'] = frozen_prob[data.index.dayofyear - 1]
        else:
            data.attrs['topo_complex'] = np.nan
            data.attrs['wetland_frac'] = np.nan
            data.attrs['porosity_gldas'] = np.nan
            data.attrs['porosity_hwsd'] = np.nan
            data['snow_prob'] = np.nan
            data['frozen_prob'] = np.nan

        if absolute_sm:
            # no error assumed for porosity values, i.e. variance = 0
            por_var = 0.

            data['abs_sm_gldas'] = data['sm'] / \
                100.0 * data.attrs['porosity_gldas']
            data['abs_sm_noise_gldas'] = np.sqrt(
                por_var * (data['sm'] / 100.0)**2 + data['sm_noise']**2 *
                (data.attrs['porosity_gldas'] / 100.0)**2)

            data['abs_sm_hwsd'] = data['sm'] / \
                100.0 * data.attrs['porosity_hwsd']
            data['abs_sm_noise_hwsd'] = np.sqrt(
                por_var * (data['sm'] / 100.0)**2 + data['sm_noise']**2 *
                (data.attrs['porosity_hwsd'] / 100.0)**2)
        else:
            data['abs_sm_gldas'] = np.nan
            data['abs_sm_noise_gldas'] = np.nan
            data['abs_sm_hwsd'] = np.nan
            data['abs_sm_noise_hwsd'] = np.nan

        if mask_ssf is not None:
            data = data[data['ssf'] < 2]

        if mask_frozen_prob is not None:
            data = data[data['frozen_prob'] < mask_frozen_prob]

        if mask_snow_prob is not None:
            data = data[data['snow_prob'] < mask_snow_prob]

        if (data.attrs['topo_complex'] is not None and
                data.attrs['topo_complex'] >= self.thresholds['topo_complex']):
            msg = "Topographic complexity >{:2d} ({:2d})".format(
                self.thresholds['topo_complex'], data.attrs['topo_complex'])
            warnings.warn(msg)

        if (data.attrs['wetland_frac'] is not None and
                data.attrs['wetland_frac'] >= self.thresholds['wetland_frac']):
            msg = "Wetland fraction >{:2d} ({:2d})".format(
                self.thresholds['wetland_frac'], data.attrs['wetland_frac'])
            warnings.warn(msg)

        return data



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Readers for ASCAT Level 1b data for various file formats.
"""

from pathlib import Path

from ascat.read_native.nc import AscatL1bNcFileGeneric
from ascat.read_native.hdf5 import AscatL1bHdf5FileGeneric
from ascat.read_native.bufr import AscatL1bBufrFileGeneric
from ascat.read_native.eps_native import AscatL1bEpsFileGeneric
from ascat.utils import get_file_format
from ascat.file_handling import ChronFiles


class AscatL1bFile:
    """
    Class reading ASCAT Level 1b files.
    """

    def __new__(cls, filename, file_format=None):
        """
        Return an instance of the appropriate ASCAT Level 1b file reader.

        Parameters
        ----------
        filename : str
            Filename.
        file_format : str, optional
            File format: ".nat", ".nc", ".bfr", ".h5" (default: None).
            If None file format will be guessed based on the file ending.
        """
        if file_format is None:
            if isinstance(filename, (str, Path)):
                file_format = get_file_format(filename)
            else:
                file_format = get_file_format(filename[0])

        if file_format in [".nat", ".nat.gz"]:
            return AscatL1bEpsFileGeneric(filename)
        elif file_format in [".nc", ".nc.gz"]:
            return AscatL1bNcFileGeneric(filename)
        elif file_format in [".bfr", ".bfr.gz", ".buf", "buf.gz"]:
            return AscatL1bBufrFileGeneric(filename)
        elif file_format in [".h5", ".h5.gz"]:
            return AscatL1bHdf5FileGeneric(filename)
        else:
            raise RuntimeError("ASCAT Level 1b file format unknown")



class AscatL1bBufrFileList(ChronFiles):
    """
    Class reading ASCAT L1b BUFR files.
    """

    def __init__(self, path, sat, product, filename_template=None):
        """
        Initialize.

        Parameters
        ----------
        path : str
            Path to input data.
        sat : str
            Metop satellite ("a", "b", "c").
        product : str
            Product type ("szf", "szr", "szo").
        filename_template : str, optional
            Filename template (default:
            "M0{sat}-ASCA-ASC{product}1B0200-NA-9.1-{date}.000000000Z-*-*.bfr")
        """
        sat_lut = {"a": 2, "b": 1, "c": 3}
        self.sat = sat_lut[sat]

        self.product = product.upper()

        if filename_template is None:
            filename_template = (
                "M0{sat}-ASCA-ASC{product}1B0200-NA-9.1-"
                "{date}.000000000Z-{placeholder1}-{placeholder2}.bfr")

        super().__init__(path, AscatL1bFile, filename_template, None)

    def _fmt(self, timestamp):
        """
        Definition of filename and subfolder format.

        Parameters
        ----------
        timestamp : datetime
            Time stamp.

        Returns
        -------
        fn_fmt : dict
            Filename format.
        sf_fmt : dict
            Subfolder format.
        """
        fn_read_fmt = {
            "date": timestamp.strftime("%Y%m%d%H%M%S"),
            "sat": self.sat,
            "product": self.product,
            "placeholder1": "*",
            "placeholder2": "*",
        }
        fn_write_fmt = None
        sf_read_fmt = None
        sf_write_fmt = sf_read_fmt

        return fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt


class AscatL1bNcFileList(ChronFiles):
    """
    Class reading ASCAT L1b NetCDF files.
    """

    def __init__(self, path, sat, product, filename_template=None):
        """
        Initialize.

        Parameters
        ----------
        path : str
            Path to input data.
        sat : str
            Metop satellite ("a", "b", "c").
        product : str
            Product type ("szr", "szo").
        filename_template : str, optional
            Filename template (default:
            "W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOP{sat}+ASCAT_C_EUMP_{date}_*_eps_o_{product}_l1.nc")
        """
        self.sat = sat

        lut = {"szr": "125", "szo": "250"}
        self.product = lut[product]

        if filename_template is None:
            filename_template = (
                "W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOP{sat}+"
                "ASCAT_C_EUMP_{date}_{placeholder}_eps_o_{product}_l1.nc")

        super().__init__(path, AscatL1bFile, filename_template, None)

    def _fmt(self, timestamp):
        """
        Definition of filename and subfolder format.

        Parameters
        ----------
        timestamp : datetime
            Time stamp.

        Returns
        -------
        fn_fmt : dict
            Filename format.
        sf_fmt : dict
            Subfolder format.
        """
        fn_read_fmt = {
            "date": timestamp.strftime("%Y%m%d%H%M%S"),
            "sat": self.sat.upper(),
            "product": self.product.upper(),
            "placeholder": "*",
        }
        fn_write_fmt = None
        sf_read_fmt = None
        sf_write_fmt = sf_read_fmt

        return fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt


class AscatL1bEpsFileList(ChronFiles):
    """
    Class reading ASCAT L1b Eps files.
    """

    def __init__(self, path, sat, product, filename_template=None):
        """
        Initialize.

        Parameters
        ----------
        path : str
            Path to input data.
        sat : str
            Metop satellite ("a", "b", "c").
        product : str
            Product type ("szf", "szr", "szo").
        filename_template : str, optional
            Filename template (default:
                "ASCA_{product}_1B_M0{sat}_{date}Z_*_*_*_*.nat")
        """
        sat_lut = {"a": 2, "b": 1, "c": 3, "?": "?"}
        self.sat = sat_lut[sat]
        self.product = product

        if filename_template is None:
            filename_template = "ASCA_{product}_1B_M0{sat}_{date}Z_{placeholder1}_{placeholder2}_{placeholder3}_{placeholder4}.nat"

        super().__init__(path, AscatL1bFile, filename_template, None)

    def _fmt(self, timestamp):
        """
        Definition of filename and subfolder format.

        Parameters
        ----------
        timestamp : datetime
            Time stamp.

        Returns
        -------
        fn_fmt : dict
            Filename format.
        sf_fmt : dict
            Subfolder format.
        """
        fn_read_fmt = {
            "date": timestamp.strftime("%Y%m%d%H%M%S"),
            "sat": self.sat,
            "product": self.product.upper(),
            "placeholder1": "*",
            "placeholder2": "*",
            "placeholder3": "*",
            "placeholder4": "*",
        }
        fn_write_fmt = None
        sf_read_fmt = None
        sf_write_fmt = sf_read_fmt

        return fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt


class AscatL1bHdf5FileList(ChronFiles):
    """
    Class reading ASCAT L1b HDF5 files.
    """

    def __init__(self, path, sat, product, filename_template=None):
        """
        Initialize.

        path : str
            Path to input data.
        sat : str
            Metop satellite ("a", "b", "c").
        filename_template : str, optional
            Filename template (default:
              "ASCA_SZF_1B_M0{sat}_{date}Z_*_*_*_*.h5")
        """
        sat_lut = {"a": "2", "b": "1", "c": "3", "?": "?"}
        self.sat = sat_lut[sat]
        self.product = product

        if filename_template is None:
            filename_template = "ASCA_{product}_1B_M0{sat}_{date}Z_{placeholder1}_{placeholder2}_{placeholder3}_{placeholder4}.h5"

        super().__init__(path, AscatL1bFile, filename_template, None)

    def _fmt(self, timestamp):
        """
        Definition of filename and subfolder format.

        Parameters
        ----------
        timestamp : datetime
            Time stamp.

        Returns
        -------
        fn_fmt : dict
            Filename format.
        sf_fmt : dict
            Subfolder format.
        """
        fn_read_fmt = {
            "date": timestamp.strftime("%Y%m%d%H%M%S"),
            "sat": self.sat,
            "product": self.product.upper(),
            "placeholder1": "*",
            "placeholder2": "*",
            "placeholder3": "*",
            "placeholder4": "*",
        }
        fn_write_fmt = None
        sf_read_fmt = None
        sf_write_fmt = sf_read_fmt

        return fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import tempfile

from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from flox.xarray import xarray_reduce
from dask.array import unique as da_unique

from ascat.swath import SwathGridFiles
from ascat.product_info import get_swath_product_id
from ascat.utils import dtype_to_nan


class TemporalSwathAggregator:
    """Class to aggregate ASCAT data its location ids over time."""

    def __init__(
        self,
        filepath,
        start_dt,
        end_dt,
        t_delta,
        agg,
        snow_cover_mask=80,
        frozen_soil_mask=80,
        subsurface_scattering_mask=5,
        ssm_sensitivity_mask=1,
        no_masking=False,
        sat=None,
    ):
        """Initialize the class.

        Parameters
        ----------
        filepath : str
            Path to the data.
        start_dt : str
            Start date and time (formatted e.g. 2020-01-01T00:00:00).
        end_dt : str
            End date and time (formatted e.g. 2020-02-01T00:00:00).
        t_delta : str
            Time period for aggregation (e.g. 1D, 1W, 1M, 1Y, 3M, 4Y, etc.).
        agg : str
            Aggregation method (e.g. mean, median, std).
        snow_cover_mask : int, optional
            Snow cover probability value above which to mask the source data.
        frozen_soil_mask : int, optional
            Frozen soil probability value above which to mask the source data.
        subsurface_scattering_mask : int, optional
            Subsurface scattering probability value above which to mask
            the source data.
        ssm_sensitivity_mask : float, optional
            Soil moisture sensitivity value above which to mask
            the source data.
        no_masking : boolean, optional
            Ignore all masks (default: False).
        sat: str, optional
            Regex pattern indicating which METOP satellite(s) to use, e.g. A, [BC], or '*'.
            `None` will use the product's default satellite (usually all of them).
            (default: None)
        """
        self.filepath = filepath

        fmt = "%Y-%m-%dT%H:%M:%S"
        self.start_dt = datetime.datetime.strptime(start_dt, fmt)
        self.end_dt = datetime.datetime.strptime(end_dt, fmt)
        self.timedelta = pd.Timedelta(t_delta)
        self.no_masking = no_masking

        agg_methods = [
            "mean", "median", "mode", "std", "min", "max", "argmin", "argmax",
            "quantile", "first", "last"
        ]

        if agg in agg_methods:
            agg = "nan" + agg

        self.agg = agg

        # assumes ONLY swath files are in the folder
        first_fname = str(next(Path(filepath).rglob("*.nc")).name)
        product = get_swath_product_id(first_fname)
        self.product = product

        if sat is not None:
            self.fmt_kwargs = {"sat": sat}
        else:
            self.fmt_kwargs = {}

        self.collection = SwathGridFiles.from_product_id(
            Path(filepath), product)

        self.grid = self.collection.grid
        self.agg_vars = {
            "surface_soil_moisture": {
                "dtype": np.dtype("int16"),
                "scale_factor": 1e-2,
            },
            "backscatter40": {
                "dtype": np.dtype("int32"),
                "scale_factor": 1e-7,
            },
        }
        self.mask_probs = {
            "snow_cover_probability": snow_cover_mask,
            "frozen_soil_probability": frozen_soil_mask,
            "subsurface_scattering_probability": subsurface_scattering_mask,
            "surface_soil_moisture_sensitivity": ssm_sensitivity_mask,
        }

    def _set_metadata(self, ds):
        """Add appropriate metadata to datasets."""
        return ds

    def _create_output_encoding(self):
        """Create NetCDF encoding."""

        output_encoding = {
            "latitude": {
                "dtype": np.dtype("int32"),
                "scale_factor": 1e-6,
                "zlib": True,
                "complevel": 4,
                "_FillValue": dtype_to_nan[np.dtype("int32")],
                "missing_value": dtype_to_nan[np.dtype("int32")],
            },
            "longitude": {
                "dtype": np.dtype("int32"),
                "scale_factor": 1e-6,
                "zlib": True,
                "complevel": 4,
                "_FillValue": dtype_to_nan[np.dtype("int32")],
                "missing_value": dtype_to_nan[np.dtype("int32")],
            },
            "time": {
                "dtype": np.dtype("float64"),
                "zlib": True,
                "complevel": 4,
                "_FillValue": 0,
                "missing_value": 0,
            },
        }

        for var in self.agg_vars:
            if var in output_encoding:
                continue

            output_encoding[var] = {
                "dtype": self.agg_vars[var]["dtype"],
                "scale_factor": self.agg_vars[var]["scale_factor"],
                "zlib": True,
                "complevel": 4,
                "_FillValue": dtype_to_nan[self.agg_vars[var]["dtype"]],
                "missing_value": dtype_to_nan[self.agg_vars[var]["dtype"]],
            }

        return output_encoding

    def write_time_steps(self, outpath):
        """
        Loop through time steps and write them to file.

        Parameters
        ----------
        outpath : str
            Output path.
        """
        product_id = self.product.lower().replace("_", "-")
        grid_sampling = str(self.grid.res) + "km"

        if self.agg is not None:
            datasets = self.get_aggregated_time_steps()
            agg_str = f"_{self.agg}"
        else:
            datasets = self.get_time_steps()
            agg_str = "_data"

        fmt = "%Y%m%d%H%M%S"

        paths = []
        output_encoding = self._create_output_encoding()
        print("saving datasets...", end="\r")
        for ds in datasets:
            step_start_str = (
                np.datetime64(ds.attrs["start_time"]).astype(
                    datetime.datetime).strftime(fmt))
            step_end_str = (
                np.datetime64(ds.attrs["end_time"]).astype(
                    datetime.datetime).strftime(fmt))

            out_name = (f"ascat"
                        f"_{product_id}"
                        f"_{grid_sampling}"
                        f"{agg_str}"
                        f"_{step_start_str}"
                        f"_{step_end_str}.nc")
            print("saving output")
            ds.to_netcdf(
                Path(outpath) / out_name,
                encoding=output_encoding,
            )

            paths.append(Path(outpath) / out_name)

        return paths

    def get_time_steps(self):
        """
        Loop through time steps of the range, return the merged data
        for each unmodified.
        """
        time_steps = pd.date_range(
            start=self.start_dt, end=self.end_dt, freq=self.timedelta)

        for timestep in time_steps:
            print("reading data for time step:", timestep, end="\r")
            step_start = timestep
            step_end = timestep + self.timedelta
            ds_step = self.collection.read((step_start, step_end),
                                           **self.fmt_kwargs)
            step_end = step_end - pd.Timedelta("1s")
            ds_step.attrs["start_time"] = np.datetime64(step_start).astype(
                str)
            ds_step.attrs["end_time"] = np.datetime64(step_end).astype(str)
            yield ds_step


    def get_aggregated_time_steps(self):
        """Loop through data in time steps, aggregating it over time."""
        for ds in self.get_time_steps():
            present_agg_vars = [
                var for var in self.agg_vars if var in ds.variables
            ]


            print("masking data...", end="\r")
            global_mask = (ds.surface_flag != 0)

            ds = ds.where(~global_mask, drop=False)

            if not self.no_masking:
                variable_masks = {
                    "surface_soil_moisture": (
                        (ds["frozen_soil_probability"]
                        > self.mask_probs["frozen_soil_probability"])
                        | (ds["snow_cover_probability"]
                        > self.mask_probs["snow_cover_probability"])
                        | (ds["subsurface_scattering_probability"]
                        > self.mask_probs["subsurface_scattering_probability"])
                        | (ds["surface_soil_moisture_sensitivity"]
                        < self.mask_probs["surface_soil_moisture_sensitivity"])),
                }

                for var, var_mask in variable_masks.items():
                    ds[var] = ds[var].where(~var_mask, drop=False)

            print("grouping data...           ")
            expected_location_ids = da_unique(ds["location_id"].data).compute()

            # remove NaN from the expected location ids (this was introduced by the masking)
            expected_location_ids = expected_location_ids[
                ~np.isnan(expected_location_ids)]

            # grouped_ds the data by time_steps and location_id and aggregate it
            grouped_ds = xarray_reduce(
                ds[present_agg_vars],
                ds["location_id"],
                expected_groups=(expected_location_ids,),
                func=self.agg)

            # convert the location_id back to an integer
            grouped_ds["location_id"] = grouped_ds["location_id"].astype(int)
            step_start = ds.attrs["start_time"]
            grouped_ds["time"] = np.datetime64(step_start, "ns")

            lons, lats = self.grid.gpi2lonlat(grouped_ds.location_id.values)
            grouped_ds["longitude"] = ("location_id", lons)
            grouped_ds["latitude"] = ("location_id", lats)
            grouped_ds = grouped_ds.set_coords(["longitude", "latitude"])
            grouped_ds = self._set_metadata(grouped_ds)

            yield grouped_ds



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import argparse
from pathlib import Path

import xarray as xr

from ascat.grids.grid_registry import GridRegistry
from ascat.product_info import get_swath_product_id
from ascat.product_info import swath_io_catalog
from ascat.regrid.regrid import regrid_swath_ds
from ascat.regrid.regrid import retrieve_or_store_grid_lut


def parse_args_swath_regrid(args):
    """
    Parse command line arguments for regridding an ASCAT swath file
    to a regular grid.

    Parameters
    ----------
    args : list
        Command line arguments.

    Returns
    -------
    parser : ArgumentParser
        Argument Parser object.
    """
    parser = argparse.ArgumentParser(
        description="Regrid an ASCAT swath file to a regular grid")
    parser.add_argument(
        "filepath", metavar="FILEPATH", help="Path to file or folder")
    parser.add_argument(
        "outpath", metavar="OUTPATH", help="Path to the output data")
    parser.add_argument(
        "regrid_deg",
        metavar="REGRID_DEG",
        type=float,
        help="Target grid spacing in degrees")
    parser.add_argument(
        "--product_id",
        metavar="PRODUCT_ID",
        help="Product identifier (e.g. H129, H125, H121, etc.). If not provided, an attempt is made to determine it from the file name.")
    parser.add_argument(
        "--grid_store",
        metavar="GRID_STORE",
        help="Path for storing/loading lookup tables")
    parser.add_argument(
        "--suffix",
        metavar="SUFFIX",
        help="File suffix (default: _REGRID_DEGdeg)")

    return parser.parse_args(args)


def swath_regrid_main(cli_args):
    """
    Regrid an ASCAT swath file or directory of swath files
    to a regular grid and write the results to disk.

    Parameters
    ----------
    cli_args : list
        Command line arguments.
    """
    args = parse_args_swath_regrid(cli_args)
    filepath = Path(args.filepath)
    trg_grid_size = args.regrid_deg

    outpath = Path(args.outpath)
    outpath.parent.mkdir(parents=True, exist_ok=True)

    if args.suffix:
        suffix = args.suffix
    else:
        suffix = f"_{args.regrid_deg}deg"

    if filepath.is_dir():
        files = list(filepath.glob("**/*.nc"))
    elif filepath.is_file() and filepath.suffix == ".nc":
        files = [filepath]
    else:
        raise RuntimeError("No files found at the provided filepath")

    first_file = files[0]

    if args.product_id:
        product_id = args.product_id
    else:
        try:
            product_id = get_swath_product_id(str(first_file.name))
        except ValueError:
            raise RuntimeError(
                f"Could not determine product identifier from file name {str(first_file.name)} "
                "Please provide the --product_id argument."
            )

    if product_id is None:
        raise RuntimeError(
            "Could not determine product identifier from file name. "
            "Please provide the --product_id argument."
        )

    registry = GridRegistry()
    product = swath_io_catalog[product_id]
    src_grid = registry.get(product.grid_name)
    src_grid_size = src_grid.res

    src_grid_id = f"fib_grid_{src_grid_size}km"
    trg_grid_id = f"reg_grid_{trg_grid_size}deg"

    trg_grid, grid_lut = retrieve_or_store_grid_lut(
        src_grid,
        src_grid_id,
        trg_grid_id,
        trg_grid_size,
        args.grid_store)

    for f in files:
        outfile = outpath / Path(f.stem + suffix + f.suffix)

        with xr.open_dataset(f, decode_cf=False, mask_and_scale=False) as ds:
            regrid_ds = regrid_swath_ds(ds, src_grid, trg_grid, grid_lut)
            regrid_ds.to_netcdf(outfile)


def run_swath_regrid():
    """Run command line interface for temporal aggregation of ASCAT data."""
    swath_regrid_main(sys.argv[1:])
